import uuid

from gcs_cli.helpers import (
    default_identity_id_resolver,
    filter_fields,
    requires_login,
    sorted_json_field,
)
from gcs_cli.parsing import (
    collection_id_arg,
    globus_command,
    include_private_policies_opt,
)
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.services.gcs import connector_id_to_display_name, get_client

STANDARD_FIELDS = [
    ("Display Name", "display_name"),
    ("Owner", default_identity_id_resolver.field),
    ("ID", "id"),
    ("Collection Type", "collection_type"),
    ("Storage Gateway ID", "storage_gateway_id"),
    ("Connector", lambda x: connector_id_to_display_name(x["connector_id"])),
    ("Allow Guest Collections", "allow_guest_collections"),
    ("Disable Anonymous Writes", "disable_anonymous_writes"),
    ("High Assurance", "high_assurance"),
    ("Authentication Timeout", "authentication_timeout_mins"),
    ("Multi-factor Authentication", "require_mfa"),
    ("Manager URL", "manager_url"),
    ("HTTPS URL", "https_url"),
    ("TLSFTP URL", "tlsftp_url"),
    ("Force Encryption", "force_encryption"),
    ("Public", "public"),
    ("Organization", "organization"),
    ("Department", "department"),
    ("Keywords", "keywords"),
    ("Description", "description"),
    ("Contact E-mail", "contact_email"),
    ("Contact Info", "contact_info"),
    ("Collection Info Link", "info_link"),
    ("Google Cloud Storage Project", "policies.project"),
]

PRIVATE_FIELDS = [
    ("Root Path", "root_path"),
    ("Default Directory", "default_directory"),
    ("Sharing Path Restrictions", sorted_json_field("sharing_restrict_paths")),
    ("Sharing Allowed Users", "sharing_users_allow"),
    ("Sharing Denied Users", "sharing_users_deny"),
    ("Sharing Allowed POSIX Groups", "policies.sharing_groups_allow"),
    ("Sharing Denied POSIX Groups", "policies.sharing_groups_deny"),
]


@globus_command("show", short_help="Show a Collection definition")
@include_private_policies_opt
@collection_id_arg
@requires_login
def collection_show(include_private_policies: bool, collection_id: uuid.UUID) -> None:
    """
    Show a Collection on the current Endpoint
    """
    client = get_client()

    params = {}
    fields = STANDARD_FIELDS

    if include_private_policies:
        params["include"] = "private_policies"
        fields += PRIVATE_FIELDS

    res = client.get_collection(collection_id, **params).data

    # walk the list of all known fields and reduce the rendering to only look
    # for fields which are actually present
    real_fields = filter_fields(fields, res)

    formatted_print(
        res,
        text_format=FORMAT_TEXT_RECORD,
        fields=real_fields,
    )
