import uuid
from typing import Optional, Union

import click

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import UUIDorUsername, globus_command, storage_gateway_id_arg
from gcs_cli.sdk_client.data import ActiveScaleUserCredentialPolicies
from gcs_cli.services.gcs import ACTIVESCALE_CONNECTOR_ID

from ._s3_create import _s3_create


@globus_command(
    "activescale-create",
    disable_options=["format"],
    short_help="Create an Active Scale User Credential",
)
@click.option(
    "--globus-identity",
    type=UUIDorUsername(),
    required=True,
    help="Globus identity id or username id to associate the credential with",
)
@click.option(
    "--mapped-user",
    help="GCSv5 mapped identity username. If not provided, defaults to the Globus "
    "identity username",
)
@click.option(
    "--s3-access-key-id",
    help="S3 Access Key ID. If not provided, this command will prompt for it",
    prompt=True,
    envvar="S3_ACCESS_KEY_ID",
)
@click.option(
    "--s3-secret-key",
    help="S3 secret key. If not provided, this command will prompt for it",
    prompt=True,
    envvar="S3_SECRET_KEY",
)
@click.option(
    "--replace-existing",
    is_flag=True,
    help="Replace a user credential that already exists. If not provided, existing "
    "credentials will not be updated",
)
@storage_gateway_id_arg
@requires_login
def activescale_create(
    globus_identity: Union[str, uuid.UUID],
    mapped_user: Optional[str],
    replace_existing: bool,
    s3_access_key_id: str,
    s3_secret_key: str,
    storage_gateway_id: str,
) -> None:
    """
    Create an Active Scale user credential for a particular user mapping on a
    storage gateway. By default, users can only create credentials for their
    own account mappings.  However, if the admin_managed_credentials property
    is set on a storage gateway, then endpoint administrators may create
    credentials on behalf of other users.
    """

    return _s3_create(
        globus_identity,
        mapped_user,
        replace_existing,
        s3_access_key_id,
        s3_secret_key,
        storage_gateway_id,
        ACTIVESCALE_CONNECTOR_ID,
        ActiveScaleUserCredentialPolicies,
    )
