import re
from typing import Callable, Optional, Union, cast

from OpenSSL import crypto


def load_certificate(path: str) -> str:
    """
    Load the X.509 certificate from path, returning a normalized output
    containing just the PEM-encoded data

    :param path:
        Path to the certificate file.
    """
    with open(path, "r") as f:
        data = f.read()
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, data.encode())
    return crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode()


def load_certificate_chain(path: str) -> Optional[str]:
    """
    Load a list of X.509 certificates from path, returning a normalized output
    containing just the PEM-encoded data

    :param path:
        Path to the certificate chain file.
    """
    if path:
        certs = []
        start = r"-----BEGIN CERTIFICATE-----[\r]?\n"
        end = r"-----END CERTIFICATE-----[\r]?[\n]?"
        with open(path, "r") as f:
            data = f.read()
        while data:
            startp = re.search(start, data, re.MULTILINE)
            endp = re.search(end, data, re.MULTILINE)
            if startp and endp and startp.start() < endp.end():
                cert = crypto.load_certificate(
                    crypto.FILETYPE_PEM,
                    data[startp.start() : endp.end()].encode(),
                )
                certs.append(
                    crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode()
                )
                data = data[endp.end() :]
            else:
                break
        return "\n".join(certs)
    return None


def load_private_key(
    path: str, passphrase: Optional[Union[str, Callable[..., bytes]]] = None
) -> str:
    """
    Load a private key from path, returning a normalized output
    containing just the PEM-encoded data, using the optional passphrase to
    decrypt the data if needed.

    :param path:
        Path to the private key file.
    :param passphrase:
        String or callable to return a passphrase
    """
    with open(path, "r") as f:
        data = f.read()

    pp_arg: Optional[Union[bytes, Callable[[], bytes]]] = None
    if isinstance(passphrase, str):
        pp_arg = passphrase.encode()
    else:
        pp_arg = passphrase

    private_key = crypto.load_privatekey(crypto.FILETYPE_PEM, data, passphrase=pp_arg)
    return cast(str, crypto.dump_privatekey(crypto.FILETYPE_PEM, private_key).decode())
