import os
import pathlib
import stat
from typing import cast, Dict, Iterable, List, Optional

import click
import globus_sdk

from ..models import (
    BlackPearlCredential,
    BlackPearlCredentialYamlDict,
    BlackPearlGuestCollection,
    BlackPearlMappedCollection,
    BlackPearlStorageGateway,
    Credential,
    CredentialYamlDict,
    GuestCollection,
    MappedCollection,
    MigrationPlan,
    PosixCredential,
    PosixGuestCollection,
    PosixMappedCollection,
    PosixStorageGateway,
    S3Credential,
    S3CredentialYamlDict,
    S3GuestCollection,
    S3MappedCollection,
    S3StorageGateway,
)
from ..parsers import BlackPearlConfig, GCSConfig, GridFTPOptionParser, S3Config
from ..util import get_api, AuthorizeSharing, migration_guide_url
from ..version import __version__
from .decorators import requires_logged_in


@click.command("create")
@click.option("--no-write", is_flag=True, hidden=True)
@click.version_option(version=__version__)
@requires_logged_in
def create(no_write: bool) -> MigrationPlan:
    os.umask(stat.S_IRWXG | stat.S_IRWXO)
    gcs_config = GCSConfig()
    epid = gcs_config.endpoint_id

    while not epid:
        endpoint_uuid = click.prompt("Enter GCS v4 endpoint ID", type=click.UUID)
        epid = str(endpoint_uuid)

    click.echo("Parsing endpoint state...")
    try:
        gridftp_config = GridFTPOptionParser(
            config_file="/etc/gridftp.conf", config_dir="/etc/gridftp.d"
        )
    except (FileNotFoundError, ValueError) as e:
        click.echo(f"WARNING: Unable to load gridftp configuration file: {e}", err=True)
        gridftp_config = GridFTPOptionParser()

    version_tag = gridftp_config.get_string_option("version_tag")
    if version_tag and version_tag.startswith("GCS-5"):
        raise click.ClickException(
            "GridFTP server configuration is for a GCSv5 endpoint"
        )

    migration_plan = MigrationPlan(".")
    try:
        migration_plan.gcs_config = gcs_config
        migration_plan.gridftp_config = gridftp_config
    except Exception as e:
        click.echo(
            f"Unable to parse the local configuration files: {e}.\nmigration plan data will be incomplete...",
            err=True,
        )

    api = get_api()

    res = api.get(f"/endpoint/{epid}")
    transfer_endpoint = res.data
    if transfer_endpoint.get("non_functional_endpoint_id"):
        raise click.ClickException(f"Endpoint {epid} is a GCS v5 collection")
    elif transfer_endpoint.get("gcs_manager_url"):
        raise click.ClickException(f"Endpoint {epid} is a GCS v5 endpoint")

    click.echo("Parsing endpoint roles...")
    role_list_document = None
    try:
        res = api.get(f"/endpoint/{epid}/role_list")
        role_list_document = res.data
    except globus_sdk.TransferAPIError as e:
        raise click.ClickException(e.message)

    authorization: Optional[AuthorizeSharing] = None
    if migration_plan.gsi_authz_configuration:
        authorization = AuthorizeSharing(migration_plan.gsi_authz_configuration)
    else:
        click.echo(
            "WARNING: Unable to parse GSI authorization configuration, local "
            "users for guest collections can not be determined",
            err=True,
        )

    dsi = gridftp_config.get_string_option("load_dsi_module")

    if dsi == "file":
        posix_storage_gateway = PosixStorageGateway()
        posix_storage_gateway.gridftp_config = gridftp_config
        posix_storage_gateway.display_name = (
            transfer_endpoint["display_name"] or transfer_endpoint["canonical_name"]
        )
        migration_plan.storage_gateway = posix_storage_gateway
    elif dsi == "s3":
        s3_storage_gateway = S3StorageGateway()
        s3_storage_gateway.gridftp_config = gridftp_config
        s3_storage_gateway.display_name = (
            transfer_endpoint["display_name"] or transfer_endpoint["canonical_name"]
        )
        s3_storage_gateway.s3_config = S3Config()
        migration_plan.storage_gateway = s3_storage_gateway
    elif dsi == "blackpearl":
        blackpearl_storage_gateway = BlackPearlStorageGateway()
        blackpearl_storage_gateway.gridftp_config = gridftp_config
        blackpearl_storage_gateway.display_name = (
            transfer_endpoint["display_name"] or transfer_endpoint["canonical_name"]
        )
        blackpearl_config_path = gridftp_config.get_string_option(
            "$BLACKPEARL_DSI_CONFIG_FILE"
        )
        blackpearl_storage_gateway.blackpearl_config = BlackPearlConfig(
            path=blackpearl_config_path
        )
        migration_plan.storage_gateway = blackpearl_storage_gateway

    mapped_collection: MappedCollection
    if dsi == "file":
        mapped_collection = PosixMappedCollection()
    elif dsi == "s3":
        mapped_collection = S3MappedCollection()
    elif dsi == "blackpearl":
        mapped_collection = BlackPearlMappedCollection()

    mapped_collection.gridftp_config = gridftp_config
    mapped_collection.transfer_endpoint = transfer_endpoint
    # Ignore implicit role for owner if the endpoint is unmanaged
    if transfer_endpoint.get("subscription_id"):
        mapped_collection.role_list_document = role_list_document
    migration_plan.mapped_collection = mapped_collection

    click.echo("Enumerating shared endpoints...")
    has_next_page = True
    next_page_params: Dict[str, str] = {}
    shared_endpoints = []
    while has_next_page:
        res = api.get(
            f"/endpoint/{epid}/shared_endpoint_list", query_params=next_page_params
        )
        next_token = res.get("next_token")
        has_next_page = next_token is not None
        next_page_params = {"next_token": next_token}
        shared_endpoints += [e["id"] for e in res.get("shared_endpoints", [])]

    sharing_state_dir = gridftp_config.get_string_option("sharing_state_dir")
    if not sharing_state_dir:
        sharing_state_dir = "$HOME/.globus/sharing"

    click.echo("Parsing shared endpoint state...")
    migration_plan.guest_collections = []
    with click.progressbar(
        shared_endpoints,
        width=30,
        show_eta=False,
        item_show_func=lambda x: cast(str, x),
    ) as bar:
        for share in cast(Iterable, bar):
            res = api.get(f"/endpoint/{share}")
            share_transfer_endpoint = res.data

            res = api.get(f"/endpoint_manager/endpoint/{share}/access_list")
            access_list = res.data

            res = api.get(f"/private/endpoint/{share}/anonymous_effective_permissions")
            anonymous_effective_permissions = res.data

            gc: GuestCollection
            if dsi == "file":
                gc = PosixGuestCollection()
            elif dsi == "s3":
                gc = S3GuestCollection()
            elif dsi == "blackpearl":
                gc = BlackPearlGuestCollection()

            gc.transfer_endpoint = share_transfer_endpoint
            gc.access_list = access_list
            gc.anonymous_effective_permissions = anonymous_effective_permissions
            gc.gridftp_config = gridftp_config

            res = api.get(f"/endpoint/{share}/role_list")
            role_list = res.data
            gc.role_list = role_list

            assert gc.user_cert_chain is not None

            username: Optional[str] = None
            subject: Optional[str] = None
            idx: Optional[int] = None

            if authorization:
                username, subject, idx = authorization.map_and_authorize_sharing(
                    gc.user_cert_chain
                )
                if username is None:
                    click.echo(
                        f"Unable to determine local mapping for guest collection {gc.id}",
                        err=True,
                    )

            credential_data: CredentialYamlDict = {
                "username": username,
                "subject": subject,
                "gsi_authz_callout_index": idx,
                "identity_id": None,
                "identity_username": None,
            }
            credential: Credential
            if dsi == "file":
                credential = PosixCredential(credential_data)
            elif dsi == "s3":
                s3_credential_data = cast(S3CredentialYamlDict, credential_data)
                s3_credential_data["s3_key_id"] = None
                s3_credential_data["s3_secret_key"] = None
                credential = S3Credential(s3_credential_data)
                if (
                    s3_storage_gateway.s3_mapfile_parser
                    and credential.username
                    and credential.username in s3_storage_gateway.s3_mapfile_parser
                ):
                    entry = s3_storage_gateway.s3_mapfile_parser[credential.username]
                    if entry:
                        credential.s3_key_id = entry.access_key_id
                        credential.s3_secret_key = entry.secret_key
            elif dsi == "blackpearl":
                blackpearl_credential_data = cast(
                    BlackPearlCredentialYamlDict, credential_data
                )
                blackpearl_credential_data["access_id"] = None
                blackpearl_credential_data["secret_key"] = None
                credential = BlackPearlCredential(blackpearl_credential_data)
                access_id_parser = (
                    blackpearl_storage_gateway.blackpearl_access_id_file_parser
                )
                if (
                    access_id_parser
                    and credential.username
                    and credential.username in access_id_parser
                ):
                    bp_entry = access_id_parser[credential.username]
                    if bp_entry:
                        credential.access_id = bp_entry.access_id
                        credential.secret_key = bp_entry.secret_key

            gc.credential = credential
            gc.credential_id = credential.id

            share_state_file = (
                pathlib.Path(sharing_state_dir)
                / f"share-{anonymous_effective_permissions['sharing_uuid']}"
            )

            gc.share_state_file = share_state_file

            if mapped_collection.force_encryption:
                gc.force_encryption = True

            if mapped_collection.force_verify:
                gc.force_verify = True
                gc.disable_verify = False

            migration_plan.guest_collections.append(gc)
            migration_plan.credentials.append(credential)

    if not no_write:
        unread_sharing_state: List[GuestCollection] = []
        for gc in migration_plan.guest_collections:
            if gc.base_path is None:
                unread_sharing_state.append(gc)
        if unread_sharing_state:
            click.echo(
                "Unable to read sharing state files. In most cases this is not "
                "a problem, but please read the documentation at "
                f"{migration_guide_url}#sharing-state-files "
                "to determine if you need to update your migration plan to set "
                "the collection base paths for your guest collections"
            )

        click.echo("Writing endpoint state to disk...")
        migration_plan.dump()
    return migration_plan


if __name__ == "__main__":
    create()
