from functools import wraps
from typing import Any, Callable
import click
from .login import check_logged_in
from gcs_migration.models import (
    MigrationPlan,
    MigrationPlanNotFound,
    MigrationPlanUnsupportedVersion,
)

_LOGIN_ERR_MSG = """
Before attempting any GCS v4 Migration commands, you will have to login using

  gcs-migration login
"""


def requires_logged_in(fn: Callable) -> Callable:
    @wraps(fn)
    def decorated_function(*args, **kwargs) -> Any:  # type: ignore
        if not check_logged_in():
            click.echo(_LOGIN_ERR_MSG, err=True)
            click.get_current_context().exit(1)
        return fn(*args, **kwargs)

    return decorated_function


def requires_migration_plan(fn: Callable) -> Callable:
    @wraps(fn)
    def _inner(*args, **kwargs) -> Any:  # type: ignore
        try:
            migration_plan = MigrationPlan.load()
        except MigrationPlanNotFound:
            click.echo(
                "No migration plan not found in the current directory. Change to the "
                "directory containing an existing plan and rerun this command.",
                err=True,
            )
            click.get_current_context().exit(1)
        except MigrationPlanUnsupportedVersion as e:
            click.echo(
                "The migration plan in the current directory was created by a "
                f"newer version of the GCS migration tools ({e.version}). "
                "Please upgrade the software and retry",
                err=True,
            )
            click.get_current_context().exit(1)
        fn(*args, migration_plan=migration_plan, **kwargs)

    return _inner
