import os
import pathlib
from typing import cast, Dict, List, Optional, TypeVar, Union

from .endpoint_metadata import EndpointMetadata, EndpointMetadataDict
from .role import Role
from ..parsers import GridFTPOptionParser, parse_restrict_paths
from ..typing.gcs_documents import (
    BlackPearlMappedCollectionDict,
    BlackPearlCollectionPoliciesDict,
    MappedCollectionDict,
    PathRestrictionsDict,
    PosixMappedCollectionDict,
    PosixCollectionPoliciesDict,
    S3MappedCollectionDict,
    S3CollectionPoliciesDict,
)
from ..typing.transfer_documents import EndpointDocumentDict, RoleListDocumentDict
from ..util import yaml, yaml_object


class MappedCollectionYamlDict(EndpointMetadataDict):
    force_verify: Optional[bool]
    disable_verify: Optional[bool]
    sharing_users_allow: Optional[List[str]]
    sharing_users_deny: Optional[List[str]]
    enable_https: Optional[bool]
    default_directory: Optional[str]
    allow_guest_collections: Optional[bool]
    sharing_restrict_paths: Optional[PathRestrictionsDict]
    force_encryption: Optional[bool]
    identity_id: Optional[str]
    user_message: Optional[str]
    user_message_link: Optional[str]
    storage_gateway_id: Optional[str]
    gridftp_config_data: Optional[Dict[str, Union[int, str, bool, None]]]


class BlackPearlMappedCollectionYamlDict(MappedCollectionYamlDict):
    pass


class PosixMappedCollectionYamlDict(MappedCollectionYamlDict):
    sharing_groups_allow: Optional[List[str]]
    sharing_groups_deny: Optional[List[str]]


class S3MappedCollectionYamlDict(MappedCollectionYamlDict):
    pass


T = TypeVar("T", bound=MappedCollectionYamlDict)
U = TypeVar("U", bound=MappedCollectionDict)


@yaml_object(yaml)
class MappedCollection(EndpointMetadata[T, U]):
    def __init__(self, data: Optional[T] = None) -> None:
        super().__init__()

        self.force_verify: Optional[bool] = None
        self.disable_verify: Optional[bool] = None
        self.sharing_users_allow: Optional[List[str]] = None
        self.sharing_users_deny: Optional[List[str]] = None
        self.sharing_groups_allow: Optional[List[str]] = None
        self.sharing_groups_deny: Optional[List[str]] = None
        self.enable_https: Optional[bool] = None
        self.default_directory: Optional[str] = None
        self.allow_guest_collections: Optional[bool] = None
        self.force_encryption: Optional[bool] = None
        self.sharing_restrict_paths: Optional[PathRestrictionsDict] = None
        self.identity_id: Optional[str] = None
        self.user_message: Optional[str] = None
        self.user_message_link: Optional[str] = None
        self.roles: Optional[List[Role]] = None
        self.role_list_document: Optional[RoleListDocumentDict] = None
        self.transfer_endpoint: Optional[EndpointDocumentDict] = None
        self._gridftp_config: Optional[GridFTPOptionParser] = None
        self.storage_gateway_id: Optional[str] = None
        self.gridftp_config_data: Optional[
            Dict[str, Union[int, str, bool, None]]
        ] = None

        if data:
            self.__setstate__(data)

    @property
    def transfer_endpoint(self) -> Optional[EndpointDocumentDict]:
        return self._transfer_endpoint

    @transfer_endpoint.setter
    def transfer_endpoint(self, value: Optional[EndpointDocumentDict]) -> None:
        self._transfer_endpoint = value

        if value:
            if self.force_verify is None:
                self.force_verify = value["force_verify"]
            if self.disable_verify is None:
                self.disable_verify = value["disable_verify"]
            if self.enable_https is None:
                self.enable_https = bool(value["subscription_id"])
            if self.identity_id is None:
                self.identity_id = value["owner_id"]
            if self.user_message is None:
                self.user_message = value["user_message"]
            if self.user_message_link is None:
                self.user_message_link = value["user_message_link"]
            if not self.force_encryption:
                self.force_encryption = value["force_encryption"]
            if not self.display_name:
                self.display_name = value["name"]

    @property
    def role_list_document(self) -> Optional[RoleListDocumentDict]:
        return self._role_list_document

    @role_list_document.setter
    def role_list_document(self, value: Optional[RoleListDocumentDict]) -> None:
        self._role_list_document = value
        roles = []
        if value and not self.roles:
            for role in value["DATA"]:
                new_role = Role()
                new_role.role_document = role
                roles.append(new_role)
            self.roles = roles

    @property
    def _gridftp_config(self) -> Optional[GridFTPOptionParser]:
        return self.__gridftp_config

    @_gridftp_config.setter
    def _gridftp_config(self, value: Optional[GridFTPOptionParser]) -> None:
        self.__gridftp_config = value
        if value:
            self.gridftp_config_data = value.options
            if self.sharing_users_allow is None:
                sharing_users_allow = value.get_string_option("sharing_users_allow")
                if sharing_users_allow:
                    self.sharing_users_allow = sharing_users_allow.split(",")
                else:
                    self.sharing_users_allow = None
            if self.sharing_users_deny is None:
                sharing_users_deny = value.get_string_option("sharing_users_deny")
                if sharing_users_deny:
                    self.sharing_users_deny = sharing_users_deny.split(",")
                else:
                    self.sharing_users_deny = None
            if self.sharing_groups_allow is None:
                sharing_groups_allow = value.get_string_option("sharing_groups_allow")
                if sharing_groups_allow:
                    self.sharing_groups_allow = sharing_groups_allow.split(",")
                else:
                    self.sharing_groups_allow = None
            if self.sharing_groups_deny is None:
                sharing_groups_deny = value.get_string_option("sharing_groups_deny")
                if sharing_groups_deny:
                    self.sharing_groups_deny = sharing_groups_deny.split(",")
                else:
                    self.sharing_groups_deny = None
            if self.sharing_restrict_paths is None:
                self.sharing_restrict_paths = parse_restrict_paths(
                    value.get_string_option("sharing_rp")
                )
            if self.default_directory is None:
                self.default_directory = value.get_string_option("default_dir")
            if self.allow_guest_collections is None:
                self.allow_guest_collections = bool(
                    value.get_string_option("sharing_dn")
                )
            if not self.force_encryption and value.get_bool_option("encrypt_data"):
                self.force_encryption = True

    @property
    def gridftp_config(self) -> Optional[GridFTPOptionParser]:
        return self._gridftp_config

    @gridftp_config.setter
    def gridftp_config(self, value: Optional[GridFTPOptionParser]) -> None:
        self._gridftp_config = value

    def dump(self, root: Union[str, os.PathLike]) -> None:
        path = pathlib.Path(root) / "mapped_collection"
        os.makedirs(path, mode=0o755)

        with open(path / "collection.yaml", "w") as f:
            yaml.dump(self, f)

        if self.roles:
            with open(path / "roles.yaml", "w") as f:
                yaml.dump(self.roles, f)

    @staticmethod
    def load(
        root: Union[str, os.PathLike], load_version: str
    ) -> Optional["MappedCollection"]:
        path = pathlib.Path(root) / "mapped_collection"
        collection = path / "collection.yaml"
        obj: Optional[MappedCollection] = None
        role_objs: List[Role] = []

        if os.path.exists(collection):
            with open(collection, "r") as f:
                obj = yaml.load(f)
            if isinstance(obj, PreConnectorMappedCollection):
                obj = PosixMappedCollection(obj.__getstate__())
            assert isinstance(obj, MappedCollection)

            roles = path / "roles.yaml"
            if os.path.exists(roles):
                with open(roles, "r") as f:
                    role_objs = yaml.load(f)

            obj.roles = role_objs
        return obj

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MappedCollection):
            return NotImplemented

        return self is other or self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> T:
        base_state = cast(T, super().__getstate__())
        base_state["force_verify"] = self.force_verify
        base_state["disable_verify"] = self.disable_verify
        base_state["sharing_users_allow"] = self.sharing_users_allow
        base_state["sharing_users_deny"] = self.sharing_users_deny
        base_state["enable_https"] = self.enable_https
        base_state["default_directory"] = self.default_directory
        base_state["allow_guest_collections"] = self.allow_guest_collections
        base_state["sharing_restrict_paths"] = self.sharing_restrict_paths
        base_state["force_encryption"] = self.force_encryption
        base_state["identity_id"] = self.identity_id
        base_state["storage_gateway_id"] = self.storage_gateway_id
        base_state["user_message"] = self.user_message
        base_state["user_message_link"] = self.user_message_link
        base_state["gridftp_config_data"] = self.gridftp_config_data

        return base_state

    def __setstate__(self, state: T) -> None:
        super().__setstate__(state)
        base_state = cast(MappedCollectionYamlDict, state)

        self.force_verify = base_state["force_verify"]
        self.disable_verify = base_state["disable_verify"]
        self.sharing_users_allow = base_state["sharing_users_allow"]
        self.sharing_users_deny = base_state["sharing_users_deny"]
        self.enable_https = base_state["enable_https"]
        self.default_directory = base_state["default_directory"]
        self.allow_guest_collections = base_state["allow_guest_collections"]
        self.sharing_restrict_paths = base_state["sharing_restrict_paths"]
        self.force_encryption = base_state["force_encryption"]
        self.identity_id = base_state["identity_id"]
        self.storage_gateway_id = base_state["storage_gateway_id"]
        self.user_message = base_state["user_message"]
        self.user_message_link = base_state["user_message_link"]
        self.gridftp_config_data = base_state["gridftp_config_data"]
        self.roles = None
        self.role_list_document = None

    def to_json(self) -> U:
        collection = cast(U, super().to_json())

        collection["DATA_TYPE"] = "collection#1.4.0"
        collection["collection_type"] = "mapped"

        # collection_base_path is always '/' for v4 non-share endpoints
        collection["collection_base_path"] = "/"

        if self.identity_id:
            collection["identity_id"] = self.identity_id

        if self.storage_gateway_id:
            collection["storage_gateway_id"] = self.storage_gateway_id

        if self.default_directory:
            collection["default_directory"] = self.default_directory

        if self.public is None:
            collection["public"] = True
        else:
            collection["public"] = self.public

        if self.force_encryption is not None:
            collection["force_encryption"] = self.force_encryption

        if self.force_verify is not None:
            collection["force_verify"] = self.force_verify

        if self.disable_verify is not None:
            collection["disable_verify"] = self.disable_verify

        if self.organization is not None:
            collection["organization"] = self.organization

        if self.department is not None:
            collection["department"] = self.department

        if self.keywords:
            collection["keywords"] = self.keywords

        if self.description:
            collection["description"] = self.description

        if self.contact_email:
            collection["contact_email"] = self.contact_email

        if self.contact_info:
            collection["contact_info"] = self.contact_info

        if self.info_link:
            collection["info_link"] = self.info_link

        if self.allow_guest_collections is not None:
            collection["allow_guest_collections"] = self.allow_guest_collections

        if self.sharing_restrict_paths:
            collection["sharing_restrict_paths"] = self.sharing_restrict_paths

        if self.enable_https is not None:
            collection["enable_https"] = self.enable_https

        if self.user_message:
            collection["user_message"] = self.user_message

        if self.user_message_link:
            collection["user_message_link"] = self.user_message_link

        if self.sharing_users_allow:
            collection["sharing_users_allow"] = self.sharing_users_allow

        if self.sharing_users_deny:
            collection["sharing_users_deny"] = self.sharing_users_deny

        if self.force_verify is not None:
            collection["force_verify"] = self.force_verify

        return collection

    def update(self, other: "MappedCollection[T, U]") -> None:  # type: ignore
        for attr in [
            "force_verify",
            "disable_verify",
            "sharing_users_allow",
            "sharing_users_deny",
            "enable_https",
            "default_directory",
            "allow_guest_collections",
            "force_encryption",
            "sharing_restrict_paths",
            "identity_id",
            "user_message",
            "user_message_link",
            "roles",
        ]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)
        super().update(other)


@yaml_object(yaml)
class BlackPearlMappedCollection(
    MappedCollection[BlackPearlMappedCollectionYamlDict, BlackPearlMappedCollectionDict]
):
    def __init__(
        self, data: Optional[BlackPearlMappedCollectionYamlDict] = None
    ) -> None:
        super().__init__()

        if data:
            self.__setstate__(data)

    def __getstate__(self) -> BlackPearlMappedCollectionYamlDict:
        base_state = super().__getstate__()

        return base_state

    def __setstate__(self, state: BlackPearlMappedCollectionYamlDict) -> None:
        super().__setstate__(state)

    def to_json(self) -> BlackPearlMappedCollectionDict:
        policies: BlackPearlCollectionPoliciesDict = {
            "DATA_TYPE": "blackpearl_collection_policies#1.0.0"
        }
        collection = cast(BlackPearlMappedCollectionDict, super().to_json())

        collection["policies"] = policies

        return collection

    def __str__(self) -> str:
        id_str = f" {self.id}" if self.id else ""
        display_name_str = f" {self.display_name}" if self.display_name else ""
        return f"<BlackPearlMappedCollection{id_str}{display_name_str}>"


@yaml_object(yaml)
class PosixMappedCollection(
    MappedCollection[PosixMappedCollectionYamlDict, PosixMappedCollectionDict]
):
    def __init__(self, data: Optional[PosixMappedCollectionYamlDict] = None) -> None:
        super().__init__()

        self.sharing_groups_allow: Optional[List[str]] = None
        self.sharing_groups_deny: Optional[List[str]] = None

        if data:
            self.__setstate__(data)

    @property
    def gridftp_config(self) -> Optional[GridFTPOptionParser]:
        return self._gridftp_config

    @gridftp_config.setter
    def gridftp_config(self, value: Optional[GridFTPOptionParser]) -> None:
        self._gridftp_config = value
        if value:
            if self.sharing_groups_allow is None:
                sharing_groups_allow = value.get_string_option("sharing_groups_allow")
                if sharing_groups_allow:
                    self.sharing_groups_allow = sharing_groups_allow.split(",")
                else:
                    self.sharing_groups_allow = None
            if self.sharing_groups_deny is None:
                sharing_groups_deny = value.get_string_option("sharing_groups_deny")
                if sharing_groups_deny:
                    self.sharing_groups_deny = sharing_groups_deny.split(",")
                else:
                    self.sharing_groups_deny = None

    def __getstate__(self) -> PosixMappedCollectionYamlDict:
        base_state = super().__getstate__()
        base_state["sharing_groups_allow"] = self.sharing_groups_allow
        base_state["sharing_groups_deny"] = self.sharing_groups_deny

        return base_state

    def __setstate__(self, state: PosixMappedCollectionYamlDict) -> None:
        super().__setstate__(state)

        self.sharing_groups_allow = state["sharing_groups_allow"]
        self.sharing_groups_deny = state["sharing_groups_deny"]

    def to_json(self) -> PosixMappedCollectionDict:
        policies: PosixCollectionPoliciesDict = {
            "DATA_TYPE": "posix_collection_policies#1.1.0"
        }
        collection = cast(PosixMappedCollectionDict, super().to_json())

        collection["policies"] = policies

        if self.sharing_groups_allow:
            policies["sharing_groups_allow"] = self.sharing_groups_allow

        if self.sharing_groups_deny:
            policies["sharing_groups_deny"] = self.sharing_groups_deny

        return collection

    def update(self, other: "PosixMappedCollection") -> None:  # type: ignore
        for attr in [
            "sharing_groups_allow",
            "sharing_groups_deny",
        ]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)
        super().update(other)

    def __str__(self) -> str:
        id_str = f" {self.id}" if self.id else ""
        display_name_str = f" {self.display_name}" if self.display_name else ""
        return f"<PosixMappedCollection{id_str}{display_name_str}>"


@yaml_object(yaml)
class S3MappedCollection(
    MappedCollection[S3MappedCollectionYamlDict, S3MappedCollectionDict]
):
    def __init__(self, data: Optional[S3MappedCollectionYamlDict] = None) -> None:
        super().__init__()

        if data:
            self.__setstate__(data)

    @property
    def gridftp_config(self) -> Optional[GridFTPOptionParser]:
        return self._gridftp_config

    @gridftp_config.setter
    def gridftp_config(self, value: Optional[GridFTPOptionParser]) -> None:
        self._gridftp_config = value

    def to_json(self) -> S3MappedCollectionDict:
        policies: S3CollectionPoliciesDict = {
            "DATA_TYPE": "s3_collection_policies#1.0.0"
        }
        collection = cast(S3MappedCollectionDict, super().to_json())

        collection["policies"] = policies

        return collection

    def __str__(self) -> str:
        id_str = f" {self.id}" if self.id else ""
        display_name_str = f" {self.display_name}" if self.display_name else ""
        return f"<S3MappedCollection{id_str}{display_name_str}>"


# Backward compatible mapped collection yaml object for deserializing
# collection.yaml from migration plans that predate multiple connector
# support
@yaml_object(yaml)
class PreConnectorMappedCollection(PosixMappedCollection):
    yaml_tag = "!MappedCollection"
