from typing import Any, Dict, List, Optional
import uuid

from flask import current_app, Blueprint, g, jsonify, request, Response, url_for
from sqlalchemy.sql.expression import false
import globus_sdk

from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    expects_json,
    paginated,
)
import globus.manager.api.errors as errors

from globus.manager.api.authentication import AuthenticationState
from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.authorization import (
    requires_credential_owner,
    requires_one_role,
    IdentitySet,
    PublicAuthorizer,
    StorageGatewayAuthorizer,
    EndpointRoleAuthorizer,
)
import globus.manager.api.models as models
from globus.manager.api.models import db, Result
from globus.manager.api.blueprints.decorators import (
    server_timing,
    db_read,
)
from globus.manager.api.schemas.api.result import result_schema
from globus.manager.api.schemas.api.user_credential import (
    user_credential_patch_schema,
    user_credential_schema,
)
from globus.manager.api.blueprints.utils import (
    create_object,
    update_object,
    patch_object,
    put_object,
)
from globus.manager.api.utils.log import log_auth_exception
from globus.manager.constants.roles import RoleType


blueprint = Blueprint("user_credentials", __name__, url_prefix="/api")


@blueprint.route("/user_credentials", methods=["GET"])
@paginated
@server_timing
@authenticated
def user_credentials_get() -> Response:
    """
    Get User Credentials

    The caller may only retrieve User Credentials which were
    created by an identity in the caller's identity set unless they are an
    administrator. If the endpoint is not public, returns ForbiddenError
    if the user has no credentials

    Authorizations:
    -   PublicAuthorizer
    -   StorageGatewayAuthorizer (if "storage_gateway" given)
    """
    storage_gateway = request.args.get("storage_gateway")
    marker = g.marker
    page_size = g.page_size

    endpoint_is_public = PublicAuthorizer().authorize_identities(g.auth_state)

    if storage_gateway:
        sg = models.StorageGateway.query.filter(
            models.StorageGateway.id == storage_gateway,
            ~models.StorageGateway.deleted,  # type: ignore
        ).one_or_none()
        if not sg:
            raise errors.NotFoundError()
        authorizer = StorageGatewayAuthorizer(sg, enforce_authentication_timeout=False)
        authorized_identities = authorizer.authorize_identities(g.auth_state)
    else:
        authorized_identities = IdentitySet()
        for i in g.auth_state.identity_set_data:
            authorized_identities.append(i)

    # include=all will return all credentials that aren't deleted from storage
    # gateways that have admin_managed_credentials=True as well as all of the
    # user's credentials, if the caller has an endpoint:administrator or
    # endpoint:owner role
    include_all = False
    if "all" in request.args.get("include", "").split(","):
        try:
            requires_one_role(
                role_names=[
                    RoleType.endpoint_administrator.value,
                    RoleType.endpoint_owner.value,
                ],
            )(lambda: True)()
            include_all = True
        except errors.ForbiddenError as e:
            current_app.logger.debug(
                f"Role authorization for user_credentials_post() failed with {e}"
            )

    # List of all credentials which are not deleted
    query = models.UserCredential.query.filter(
        models.UserCredential.deleted == false(),
    )

    if not include_all:
        query = query.filter(
            models.UserCredential.identity_id.in_(  # type: ignore
                [i["id"] for i in authorized_identities]
            )
        )
    if storage_gateway:
        query = query.filter_by(storage_gateway_id=storage_gateway)

    if marker:
        query = query.filter(models.ConfigurationObject.id >= marker.marker)

    query = query.order_by(models.ConfigurationObject.id).limit(page_size + 1)

    g.server_timing.start(db_read)
    query_results = []
    returned_rows = 0
    last_id = None
    for credential in query:
        last_id = credential.id
        returned_rows += 1

        if returned_rows > page_size:
            break

        if credential.is_deleted:
            continue

        if not (include_all and credential.storage_gateway.admin_managed_credentials):
            for i in authorized_identities:
                if i["id"] == credential.identity_id:
                    if (
                        credential.storage_gateway.map_identity(i)
                        == credential.username
                    ):
                        current_app.logger.debug(
                            "Adding credential %s to response", credential
                        )
                        query_results.append(credential)
                        break
                    else:
                        current_app.logger.debug("Ignoring credential without mapping")
                        break
        else:
            query_results.append(credential)
    g.server_timing.finish(db_read)

    # If the caller has an endpoint role, they can see all collections
    caller_has_endpoint_role = EndpointRoleAuthorizer.authorize_identities(g.auth_state)

    if query_results == [] and not endpoint_is_public and not caller_has_endpoint_role:
        raise errors.ForbiddenError()

    results = []
    for uc in query_results:
        results.append(user_credential_schema.dump(uc))

    if returned_rows > page_size:
        next_page_token = Base62Marker(marker_data=last_id)
    else:
        next_page_token = None

    res = result_schema.dump(Result(data=results, marker=next_page_token))

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    resp.headers["Server-Timing"] = g.server_timing.as_string()

    return resp


def get_user_credential_id_from_path() -> Optional[models.UserCredential]:
    """
    Resolves the user_credential_id view parameter to a UserCredential if
    a UserCredential with that ID is in the database.
    """
    user_credential_id = request.view_args.get("user_credential_id")

    assert user_credential_id is not None

    user_credential = None
    if user_credential_id:
        user_credential = models.UserCredential.query.filter(
            models.UserCredential.id == str(user_credential_id)
        ).one_or_none()
        if user_credential and user_credential.is_deleted:
            user_credential = None

    g.user_credential = user_credential
    return user_credential


@blueprint.route("/user_credentials/<uuid:user_credential_id>", methods=["GET"])
@authenticated
@requires_credential_owner(
    credential_selector=get_user_credential_id_from_path,
    allow_admin_if_configured=True,
)
def user_credentials_get_by_id(user_credential_id: uuid.UUID) -> Response:
    """
    Get a User Credential by its id.

    The caller must have the identity_id of the user credential
    in its identity set.

    Authorizations:
    -   CredentialOwnerAuthorizer
    """
    uc = g.user_credential

    result = user_credential_schema.dump(uc)

    res = result_schema.dump(Result(data=[result]))

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]

    return resp


def get_storage_gateway_from_credential_doc() -> Optional[models.StorageGateway]:
    storage_gateway_id = g.validated_data.get("storage_gateway_id")
    if storage_gateway_id:
        storage_gateway = models.StorageGateway.query.filter(
            models.StorageGateway.id == str(storage_gateway_id)
        ).one_or_none()
        g.storage_gateway = storage_gateway
        return storage_gateway
    return None


@blueprint.route("/user_credentials/<uuid:user_credential_id>", methods=["PATCH"])
@authenticated
@expects_json(schema=user_credential_patch_schema)
@requires_credential_owner(
    credential_selector=get_user_credential_id_from_path,
    allow_admin_if_configured=True,
)
def user_credentials_patch(user_credential_id: uuid.UUID) -> Response:
    """
    Update a user credential on a Storage Gateway. This is required
    for some connectors that require a local user name or other
    credential information to access the storage system. See the
    connector-specific documentation for details on what is needed.

    The caller must have the identity_id of the user credential
    in its identity set.

    Authorizations:
    -   CredentialOwnerAuthorizer
    """
    uc = g.user_credential

    with current_app.appsync_client.pause_notifications():
        if uc.is_deleted:
            current_app.logger.error("PATCH %s: already deleted", str(uc.id))
            raise errors.NotFoundError()
        elif g.validated_data.get("deleted"):
            raise errors.ValidationError("Cannot change deleted via PATCH")

        patch_object(uc, g.validated_data)
        res = result_schema.dump(
            Result(
                message="Updated User Credential {}".format(str(uc.id)),
                data=[user_credential_schema.dump(uc)],
            )
        )

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    g.updated_objects.append(uc.id)

    return resp


@blueprint.route("/user_credentials/<uuid:user_credential_id>", methods=["PUT"])
@authenticated
@expects_json(schema=user_credential_schema)
@requires_credential_owner(
    credential_selector=get_user_credential_id_from_path,
    allow_admin_if_configured=True,
)
def user_credentials_put(user_credential_id: uuid.UUID) -> Response:
    """
    Update a user credential on a Storage Gateway. This is required
    for some connectors that require a local user name or other
    credential information to access the storage system. See the
    connector-specific documentation for details on what is needed.

    The caller must have the identity_id of the user credential
    in its identity set.

    Authorizations:
    -   CredentialOwnerAuthorizer
    """
    uc = g.user_credential

    with current_app.appsync_client.pause_notifications():
        if uc.is_deleted:
            current_app.logger.error("PUT %s: already deleted", str(uc.id))
            raise errors.NotFoundError()
        elif g.validated_data.get("deleted", False) != g.user_credential.deleted:
            raise errors.ValidationError("Cannot change deleted via PUT")

        put_object(uc, g.validated_data)
        res = result_schema.dump(
            Result(
                message="Updated User Credential {}".format(str(uc.id)),
                data=[user_credential_schema.dump(uc)],
            )
        )

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    g.updated_objects.append(uc.id)

    return resp


@blueprint.route("/user_credentials", methods=["POST"])
@authenticated
@expects_json(schema=user_credential_schema)
def user_credentials_post() -> Response:
    """
    Create a user credential on a Storage Gateway. This is required
    for some connectors that require a local user name or other
    credential information to access the storage system. See the
    connector-specific documentation for details on what is needed.

    The caller is authorized based on the StorageGateway identity
    policies, so users with no assigned roles on the endpoint may
    be permitted to access this operation.

    Authorizations:
    -   StorageGatewayAuthorizer
    -   RoleAuthorizer
    """
    g.storage_gateway = get_storage_gateway_from_credential_doc()
    if g.storage_gateway is None:
        raise errors.BadRequestError()

    assert g.validated_data is not None

    has_role = False
    meets_gateway_policy = False

    if g.storage_gateway.admin_managed_credentials:
        # Only permit role authorization if allowed by policy
        try:
            has_role = requires_one_role(
                storage_gateway_selector=lambda: g.storage_gateway,
                role_names=[
                    RoleType.endpoint_administrator.value,
                    RoleType.endpoint_owner.value,
                ],
            )(lambda: True)()
        except errors.ForbiddenError as e:
            current_app.logger.debug(
                f"Role authorization for user_credentials_post() failed with {e}"
            )

    # Check if ids in the caller's ID set are allowed by storage gateway
    # policies.
    storage_gateway_authorizer = StorageGatewayAuthorizer(g.storage_gateway)
    sg_auth_result = storage_gateway_authorizer.authorize_identities(g.auth_state)
    if sg_auth_result:
        meets_gateway_policy = True
    elif not has_role:
        detail = None
        if sg_auth_result.error_details:
            detail = sg_auth_result.error_details[0]
        raise errors.AuthorizationError(detail=detail)

    assert has_role or meets_gateway_policy

    # identities which are authorized by storage gateway policy, indexed by the id
    authorized_identities: Dict[str, Dict[str, Any]] = {}

    if meets_gateway_policy:
        authorized_identities = {i["id"]: i for i in sg_auth_result}

    if "identity_id" not in g.validated_data:
        # Infer identity_id if not present. This can only be done if the caller
        # has a single identity in their identity set which meets the storage
        # gateway policy.
        if not meets_gateway_policy:
            raise errors.ValidationError(
                error_description="Missing identity_id property"
            )

        if len(authorized_identities) == 1:
            g.validated_data["identity_id"] = list(authorized_identities.keys())[0]
        else:
            raise errors.ValidationError(
                "Missing identity_id property, and you have more than one "
                "authorized Globus account."
            )
    assert g.validated_data["identity_id"] is not None
    requested_identity_id: str = str(g.validated_data["identity_id"])

    # identity document for the requested_identity_id value.
    authorized_identity: Optional[Dict[str, Any]] = None

    # The caller must either have the requested identity in their identity
    # set or have the endpoint role. In the latter case, we'll use the auth
    # client to resolve the identity_id to an identity dict and then validate
    # that identity against the storage gateway policies
    if requested_identity_id in authorized_identities.keys() and meets_gateway_policy:
        authorized_identity = authorized_identities[requested_identity_id]
    elif has_role:
        try:
            # Determine if the requested identity meets storage gateway policy
            # if we are authorizing because of a role. First we look up the
            # requested identity in auth and then pass it to the
            # storage_gateway_authorizer using a synthesized
            resolved_identities: List[Dict[str, Any]] = [
                i
                for i in current_app.auth_client.get_identities(
                    ids=requested_identity_id
                )
                if i["id"] == requested_identity_id
            ]
            auth_state = AuthenticationState(identity_set_data=resolved_identities)
            # New authorizer that does not enforce timeouts, since we don't
            # have any authentication information related to the identity
            # authorized by the role
            storage_gateway_authorizer = StorageGatewayAuthorizer(
                g.storage_gateway, enforce_authentication_timeout=False
            )
            sg_auth_result = storage_gateway_authorizer.authorize_identities(auth_state)
            if not sg_auth_result:
                detail = None
                if sg_auth_result.error_details:
                    detail = sg_auth_result.error_details[0]
                raise errors.AuthorizationError(detail=detail)
            authorized_identity = sg_auth_result[0]
        except globus_sdk.GlobusAPIError as e:
            log_auth_exception(e)
    else:
        raise errors.IdNotInIdentitySetError(id=requested_identity_id)

    if not authorized_identity:
        raise errors.BadRequestError(f"Invalid identity_id {requested_identity_id}")

    mapped_to = g.storage_gateway.map_identity(authorized_identity)

    if not mapped_to:
        raise errors.InvalidUserError()

    requested_username = g.validated_data.get("username")
    if requested_username is None:
        g.validated_data["username"] = mapped_to

    if requested_username and requested_username != mapped_to:
        raise errors.InvalidUserError(usernames=[mapped_to])
    if not g.storage_gateway.check_user_allowed(mapped_to):
        raise errors.InvalidUserError(usernames=[mapped_to])

    with current_app.appsync_client.pause_notifications():
        try:
            cls = models.UserCredential.polymorphic_class(
                g.storage_gateway.connector_id
            )
            old_uc = cls.query.filter(
                cls.storage_gateway_id == g.storage_gateway.id,
                cls.username == g.validated_data["username"],
                cls.identity_id == requested_identity_id,
            ).one_or_none()

            if old_uc:
                if not old_uc.deleted:
                    raise errors.ResourceExistsError(resource_ids=[old_uc.id])
            elif not old_uc:
                uc = cls(doc=g.validated_data, storage_gateway=g.storage_gateway)
        except ValueError as e:
            current_app.logger.error("Error creating using credential: {}".format(e))
            raise errors.ValidationError()

        if old_uc:
            old_uc.deleted = False
            put_object(old_uc, g.validated_data)
            uc = old_uc
        else:
            create_object(uc)

        credential_json = user_credential_schema.dump(uc)

        res = result_schema.dump(
            Result(
                http_response_code=201,
                message="Created User Credential {}".format(uc.id),
                data=[credential_json],
            )
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    response.headers["Location"] = url_for(
        ".user_credentials_get_by_id",
        user_credential_id=uc.id,
        _external=True,
        _scheme="https",
    )
    g.created_objects.append(uc.id)

    return response


@blueprint.route("/user_credentials/<uuid:user_credential_id>", methods=["DELETE"])
@authenticated
@requires_credential_owner(
    credential_selector=get_user_credential_id_from_path, allow_admin_if_configured=True
)
def user_credentials_delete(user_credential_id: uuid.UUID) -> Response:
    """
    Delete a User Credential by its id.

    The caller must have the identity_id of the user credential in its identity
    set.

    Authorizations:
    -   CredentialOwnerAuthorizer
    """
    uc = g.user_credential

    with current_app.appsync_client.pause_notifications():
        if uc.is_deleted:
            current_app.logger.error("DELETE %s: already deleted", str(uc.id))
            raise errors.NotFoundError()

        existing_collections = [
            uc.id
            for c in db.session.query(models.Collection.id)
            .filter(models.Collection.user_credential_id == uc.id)
            .all()
        ]

        if existing_collections:
            current_app.logger.info(
                "DELETE %s: in use by %d Collection(s)",
                uc.id,
                len(existing_collections),
            )

        def delete_cred(credential: models.UserCredential) -> None:
            credential.delete()

        update_object(uc, delete_cred)

        res = result_schema.dump(
            Result(
                message="Deleted User Credential {}".format(str(user_credential_id)),
                data=[],
            )
        )

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    g.deleted_objects.append(uc.id)

    return resp
