from typing import Optional
from .collection import Collection
from .storage_gateway import StorageGateway

import globus.manager.api.errors as errors
from globus.manager.constants import ConfigurationObjectTag, CollectionType


class GoogleCloudStorageCollection(Collection):
    POLICY_DATA_TYPE = "google_cloud_storage_collection_policies#1.0.0"
    POLICY_DATA_TYPES = [POLICY_DATA_TYPE]

    CONNECTOR_POLICIES = {POLICY_DATA_TYPE: ["project"]}

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.COLLECTION.name
            + "56366b96-ac98-11e9-abac-9cb6d0d9fd63"
        )
    }

    def __init__(
        self,
        appsync_data=None,
        doc=None,
        storage_gateway=None,
        mapped_collection=None,
        user_credential=None,
        endpoint=None,
    ):
        if doc:
            policies = doc.get("policies")
            if not policies:
                policies = {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                }

            project = policies.get("project")
            collection_type = CollectionType(doc.get("collection_type"))

            if collection_type == CollectionType.guest:
                if project is None:
                    project = mapped_collection.project
                    policies = {**policies, **{"project": project}}
                if not user_credential:
                    raise ValueError("user_credential")

                # Raises exception if not possible
                user_credential.can_access_project(project)
            elif project is None:
                # If the storage gateway defines only one project, use it as
                # the default if none is specified
                allowed_projects = getattr(storage_gateway, "projects", [])
                if len(allowed_projects) == 1:
                    project = allowed_projects[0]
                    policies = {**policies, **{"project": project}}
                if project is None:
                    raise ValueError(
                        "The policies document for this collection is missing the "
                        "project property. This is required to create a Google "
                        "Cloud Storage collection unless the storage gateway "
                        "is restricted to a single project. See the "
                        "documentation at https://docs.globus.org/"
                        "premium-storage-connectors/v5.4/google-cloud-storage/"
                        "#collection_policies for more details."
                    )

            # Raises an exception if the project property doesn't meet storage
            # gateway policies
            self._check_project(project, storage_gateway, mapped_collection)

            doc = {**doc, "policies": {**policies}}

        super().__init__(
            appsync_data=appsync_data,
            doc=doc,
            storage_gateway=storage_gateway,
            mapped_collection=mapped_collection,
            user_credential=user_credential,
            endpoint=endpoint,
        )

    @property
    def connector_id(self):
        """
        UUID of the Google Cloud Storage connector
        """
        return "56366b96-ac98-11e9-abac-9cb6d0d9fd63"

    @connector_id.setter
    def connector_id(self, value):
        pass

    @property
    def project(self):
        """
        Project associated with this collection. All users of this collection
        must have Google credentials which are authorized to use this account,
        and all access to Google Cloud Storage buckets will be charged to this
        account.
        """
        if self.is_guest_collection:
            return self.mapped_collection.project
        else:
            return self.get_policy("project")

    @project.setter
    def project(self, value):
        if self.is_mapped_collection:
            self.set_policy("project", value)

    @project.deleter
    def project(self):
        if self.is_mapped_collection:
            raise ValueError("Unable to delete the project property")

    def put(self, **doc):
        """
        Overload of put() that checks policies for validity. Check here and in
        check_user_credential instead of in policy setters so that object loads
        aren't broken when a collection is in an inconsistent state.
        """
        policies = doc.get("policies", {})
        if policies and policies.get("DATA_TYPE") != self.POLICY_DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        project = policies.get("project")

        if not project:
            if self.is_mapped_collection:
                raise ValueError("Google Cloud Storage collections require a project")
            else:
                project = self.mapped_collection.project

        self._check_project(
            project,
            self.storage_gateway,
            self.mapped_collection if self.is_guest_collection else None,
        )

        super().put(**doc)

    def patch(self, **doc):
        """
        Overload of patch() that checks policies for validity. Check here and in
        check_user_credential instead of in policy setters so that object loads
        aren't broken when a collection is in an inconsistent state.
        """
        policies = doc.get("policies", {})
        if policies and policies.get("DATA_TYPE") != self.POLICY_DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        if "project" in policies:
            project = policies.get("project")
            self._check_project(
                project,
                self.storage_gateway,
                self.mapped_collection if self.is_guest_collection else None,
            )

        super().patch(**doc)

    def _check_project(
        self,
        project: str,
        storage_gateway: StorageGateway,
        mapped_collection: Optional[Collection] = None,
    ) -> None:
        """
        Check that the project requested by the user creating or
        updating a collection makes sense. This will check that the
        """
        allowed_projects = getattr(storage_gateway, "projects", None)

        if allowed_projects and project not in allowed_projects:
            raise ValueError(
                "Project {} is not allowed by the "
                "policies of storage gateway {}".format(project, storage_gateway.id)
            )
        if mapped_collection and project != mapped_collection.project:
            raise ValueError(
                "Project {} is not the same as the value used by the "
                "policies of mapped collection {}".format(project, mapped_collection.id)
            )

    def check_user_credential(self, credential):
        project = self.project
        try:
            self._check_project(
                project,
                self.storage_gateway,
                self.mapped_collection if self.is_guest_collection else None,
            )
        except ValueError as v:
            raise errors.ForbiddenError(str(v))
        credential.can_access_project(project)

    def __repr__(self):
        return "<GoogleCloudStorageCollection {}>".format(self.id)
