import ipaddress
from typing import Any, Optional, Iterable, List, Mapping, Tuple, Union
import uuid

from globus.manager.constants import NodeStatus


class Node:
    """
    Model for a GCS Node.

    Each Node object in the API consists of:
      - [required] an ID of type UUID
      - [required] one or more IP addresses
      - [optional] a status field of type NodeStatus, defaults to active
      - [optional] incoming port range
      - [optional] outgoing port range
      - [optional] data interface
    """

    _DATA_TYPE = "node#1.1.0"
    _DATA_TYPES = ["node#1.0.0", "node#1.1.0"]

    def __init__(
        self,
        DATA_TYPE: Optional[str] = None,
        ip_addresses: Optional[Iterable[str]] = None,
        id: Optional[str] = None,
        status: Optional[NodeStatus] = None,
        incoming_port_range: Optional[Tuple[int, int]] = None,
        outgoing_port_range: Optional[Tuple[int, int]] = None,
        data_interface: Optional[str] = None,
    ) -> None:
        if not DATA_TYPE or not DATA_TYPE.startswith("node#1."):
            raise ValueError("Invalid DATA_TYPE for node")

        if id is None:
            self._id = str(uuid.uuid4())
        else:
            self._id = id

        if ip_addresses is None:
            raise ValueError("The ip_addresses property for a node may not be empty")

        # https://github.com/python/mypy/issues/3004
        self.ip_addresses = [ip for ip in ip_addresses]  # type: ignore

        if not status:
            self.status = NodeStatus.active
        elif isinstance(status, NodeStatus):
            self.status = status
        else:
            try:
                self.status = getattr(NodeStatus, status)
            except AttributeError:
                raise ValueError("Invalid node status {}".format(status))

        self._incoming_port_range: Optional[Tuple[int, int]] = None
        self._outgoing_port_range: Optional[Tuple[int, int]] = None
        self._data_interface: Optional[
            Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
        ] = None

        if incoming_port_range:
            self.incoming_port_range = incoming_port_range
        if outgoing_port_range:
            self.outgoing_port_range = outgoing_port_range
        if data_interface:
            self.data_interface = ipaddress.ip_address(data_interface)

    @property
    def id(self) -> str:
        return self._id

    @property
    def ip_addresses(self) -> List[Union[ipaddress.IPv4Address, ipaddress.IPv6Address]]:
        return self._ip_addresses

    @ip_addresses.setter
    def ip_addresses(
        self, value: List[Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address]]
    ) -> None:
        if value is None or len(value) == 0:
            raise ValueError("Invalid ip_addresses")
        self._ip_addresses = [ipaddress.ip_address(ip) for ip in value]

    @property
    def incoming_port_range(self) -> Tuple[int, int]:
        if not self._incoming_port_range:
            raise AttributeError(
                "The incoming_port_range property is not defined for this node"
            )
        return self._incoming_port_range

    @incoming_port_range.setter
    def incoming_port_range(
        self, value: Optional[Union[Tuple[int, int], List[int]]]
    ) -> None:
        if value:
            self._incoming_port_range = self._parse_port_range(value)
        else:
            self._incoming_port_range = None

    @property
    def outgoing_port_range(self) -> Tuple[int, int]:
        if not self._outgoing_port_range:
            raise AttributeError(
                "The outgoing_port_range property is not defined for this node"
            )
        return self._outgoing_port_range

    @outgoing_port_range.setter
    def outgoing_port_range(
        self, value: Optional[Union[Tuple[int, int], List[int]]]
    ) -> None:
        if value:
            self._outgoing_port_range = self._parse_port_range(value)
        else:
            self._outgoing_port_range = None

    @property
    def status(self) -> NodeStatus:
        return self._status

    @status.setter
    def status(self, value: Union[str, NodeStatus]) -> None:
        self._status = NodeStatus(value)

    @property
    def data_interface(self) -> Union[ipaddress.IPv4Address, ipaddress.IPv6Address]:
        if not self._data_interface:
            raise AttributeError(
                "The data_interface property is not defined for this node"
            )
        return self._data_interface

    @data_interface.setter
    def data_interface(
        self, value: Optional[Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address]]
    ) -> None:
        self._data_interface = None
        if value is not None:
            self._data_interface = ipaddress.ip_address(value)

    def put(self, node: Union["Node", Mapping[str, Any]]) -> bool:
        changed = False

        incoming_port_range: Optional[Tuple[int, int]] = None
        outgoing_port_range: Optional[Tuple[int, int]] = None
        data_interface: Optional[Union[ipaddress.IPv4Address, ipaddress.IPv6Address]]

        if not isinstance(node, Node):
            if node.get("DATA_TYPE") not in Node._DATA_TYPES:
                raise ValueError("Unknown DATA_TYPE")

            node_id_str = str(node["id"])
            ip_addresses = sorted(
                [ipaddress.ip_address(ip) for ip in node["ip_addresses"]]
            )
            status = NodeStatus(node["status"])
            pr = node.get("incoming_port_range")
            if pr:
                incoming_port_range = self._parse_port_range(pr)
            pr = node.get("outgoing_port_range")
            if pr:
                outgoing_port_range = self._parse_port_range(pr)
            data_interface = node.get("data_interface")
        else:
            node_id_str = str(node.id)
            ip_addresses = sorted(node.ip_addresses)
            status = NodeStatus(node.status)
            incoming_port_range = getattr(node, "incoming_port_range", None)
            outgoing_port_range = getattr(node, "outgoing_port_range", None)
            di = getattr(node, "data_interface", None)
            if di:
                data_interface = ipaddress.ip_address(di)

        if node_id_str != self._id:
            raise ValueError("The id property is immutable")

        if ip_addresses != self._ip_addresses:
            self.ip_addresses = ip_addresses
            changed = True

        if status != self._status:
            self._status = status
            changed = True

        if incoming_port_range != self._incoming_port_range:
            self._incoming_port_range = incoming_port_range
            changed = True

        if outgoing_port_range != self._outgoing_port_range:
            self._outgoing_port_range = outgoing_port_range
            changed = True

        if data_interface != self._data_interface:
            self._data_interface = data_interface
            changed = True

        return changed

    def patch(self, node_changes: Mapping[str, Any]) -> bool:
        changed = False
        for n in node_changes:
            if n == "DATA_TYPE":
                if node_changes[n] not in Node._DATA_TYPES:
                    raise ValueError("Unknown DATA_TYPE")
            elif n == "id":
                if str(node_changes[n]) != self.id:
                    raise ValueError("id is an immutable property")
            else:
                old_value = getattr(self, n, None)
                new_value = node_changes[n]
                if old_value != new_value:
                    setattr(self, n, new_value)
                    changed = True
        return changed

    @staticmethod
    def _parse_port_range(value: Any) -> Tuple[int, int]:
        if (
            isinstance(value, (tuple, list))
            and isinstance(value[0], int)
            and isinstance(value[1], int)
            and len(value) == 2
        ):
            start, end = (value[0], value[1])
            if (
                start < end
                and start > 1
                and start < 2**16 - 1
                and end > 1
                and end < 2**16 - 1
            ):
                return (start, end)
        raise ValueError(f"Invalid port range {value}")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Node):
            return False
        return (
            self._id == other._id
            and self._ip_addresses == other._ip_addresses
            and self._status == other._status
            and self._incoming_port_range == other._incoming_port_range
            and self._outgoing_port_range == other._outgoing_port_range
            and self._data_interface == other._data_interface
        )

    def __repr__(self) -> str:
        return "<Node {}>".format(self.id)
