from typing import List

import marshmallow as ma
from marshmallow_oneofschema import OneOfSchema

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


BOX_CONNECTOR_ID = "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63"


class BoxAppAuth(ma.Schema):
    publicKeyID = ma.fields.Str(required=True)
    privateKey = ma.fields.Str(required=True)
    passphrase = ma.fields.Str(required=True)


class BoxAppSettingsSchema(ma.Schema):
    clientID = ma.fields.Str(required=True)
    clientSecret = ma.fields.Str(required=True)
    appAuth = ma.fields.Nested(BoxAppAuth, required=True)


class BoxJsonConfigSchema(ma.Schema):
    boxAppSettings = ma.fields.Nested(BoxAppSettingsSchema, required=True)
    enterpriseID = ma.fields.Str(required=True)
    private = ["boxAppSettings", "enterpriseID"]


class BoxStoragePoliciesSchema_1_0_0(BoxJsonConfigSchema):
    _DATA_TYPE = "box_storage_policies#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    connector_id = BOX_CONNECTOR_ID


class BoxStoragePoliciesSchema_1_1_0(ma.Schema):
    _DATA_TYPE = "box_storage_policies#1.1.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    boxAppSettings = ma.fields.Nested(BoxAppSettingsSchema, required=False)
    enterpriseID = ma.fields.Str(required=False)
    client_id = ma.fields.Str(required=False)
    secret = ma.fields.Str(required=False)
    user_api_rate_limit = ma.fields.Int(required=False)
    user_credential_required = ma.fields.Bool(dump_only=True)
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        doc="URL of the auth callback that must be set on the Box developer "
        "console for the Box application of client_id.",
    )

    private = [
        "client_id",
        "secret",
        "user_api_rate_limit",
        "boxAppSettings",
        "enterpriseID",
    ]
    connector_id = BOX_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data

    @ma.validates_schema
    def old_or_new(self, data, **kwargs):
        # Either (boxAppSettings, enterpriseID) or (client_id, secret)
        partial = kwargs.get("partial", False)
        if partial:
            old_style = "boxAppSettings" in data and "enterpriseID" in data
            new_style = "client_id" in data and "secret" in data
            if not old_style and not new_style:
                return
        else:
            old_style = "boxAppSettings" in data or "enterpriseID" in data
            new_style = "client_id" in data or "secret" in data

        if not (old_style ^ new_style):
            raise ma.ValidationError(
                f"The {self._DATA_TYPE} document must contain either "
                f"boxAppSettings and enterpriseID or client_id and secret {kwargs}"
            )


box_storage_policies_1_0_0_schema = BoxStoragePoliciesSchema_1_0_0(
    load_only=BoxStoragePoliciesSchema_1_0_0.private
)
box_storage_policies_1_1_0_schema = BoxStoragePoliciesSchema_1_1_0(
    load_only=BoxStoragePoliciesSchema_1_1_0.private
)
box_storage_policies_1_0_0_private_schema = BoxStoragePoliciesSchema_1_0_0()
box_storage_policies_1_1_0_private_schema = BoxStoragePoliciesSchema_1_1_0()


class BoxStoragePoliciesSchema(OneOfSchema):
    _DATA_TYPE = box_storage_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "box_storage_policies#1.0.0": box_storage_policies_1_0_0_schema,
        "box_storage_policies#1.1.0": box_storage_policies_1_1_0_schema,
    }
    type_field_remove = False
    connector_id = BOX_CONNECTOR_ID

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    private = box_storage_policies_1_1_0_schema.private


class BoxStoragePoliciesPrivateSchema(OneOfSchema):
    _DATA_TYPE = box_storage_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "box_storage_policies#1.0.0": box_storage_policies_1_0_0_private_schema,
        "box_storage_policies#1.1.0": box_storage_policies_1_1_0_private_schema,
    }
    type_field_remove = False
    connector_id = BOX_CONNECTOR_ID

    def get_obj_type(self, obj):
        return self._DATA_TYPE


class BoxCollectionPoliciesSchema(ma.Schema):
    _DATA_TYPE = "box_collection_policies#1.0.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    private = []  # type: List[str]
    connector_id = BOX_CONNECTOR_ID


class BoxUserCredentialPoliciesSchema_1_0_0(ma.Schema):
    _DATA_TYPE = "box_user_credential_policies#1.0.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    private = []  # type: List[str]
    connector_id = BOX_CONNECTOR_ID


class BoxUserCredentialPoliciesSchema_1_1_0(ma.Schema):
    _DATA_TYPE = "box_user_credential_policies#1.1.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    sub = ma.fields.Str(allow_none=True)
    email = ma.fields.Email(allow_none=True)
    access_token = ma.fields.Str(required=False, allow_none=True)
    refresh_token = ma.fields.Str(required=False, allow_none=True)
    token_expiry = ma.fields.AwareDateTime()
    scopes = ma.fields.List(ma.fields.Str)
    max_upload = ma.fields.Int()

    private = ["access_token", "refresh_token", "max_upload"]
    connector_id = BOX_CONNECTOR_ID


box_storage_policies_private_schema = BoxStoragePoliciesPrivateSchema()
box_storage_policies_schema = BoxStoragePoliciesSchema(
    load_only=BoxStoragePoliciesSchema.private
)


box_user_credential_policies_1_0_0_schema = BoxUserCredentialPoliciesSchema_1_0_0(
    load_only=BoxUserCredentialPoliciesSchema_1_0_0.private
)
box_user_credential_policies_1_1_0_schema = BoxUserCredentialPoliciesSchema_1_1_0(
    load_only=BoxUserCredentialPoliciesSchema_1_1_0.private
)
box_user_credential_policies_1_0_0_private_schema = (
    BoxUserCredentialPoliciesSchema_1_0_0()
)
box_user_credential_policies_1_1_0_private_schema = (
    BoxUserCredentialPoliciesSchema_1_1_0()
)
box_collection_policies_schema = BoxCollectionPoliciesSchema(
    load_only=BoxCollectionPoliciesSchema.private
)
box_collection_policies_private_schema = BoxCollectionPoliciesSchema()


class BoxUserCredentialPoliciesSchema(OneOfSchema):
    _DATA_TYPE = box_user_credential_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "box_user_credential_policies#1.0.0": box_user_credential_policies_1_0_0_schema,
        "box_user_credential_policies#1.1.0": box_user_credential_policies_1_1_0_schema,
    }
    type_field_remove = False
    connector_id = BOX_CONNECTOR_ID

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    private = box_user_credential_policies_1_1_0_schema.private + []


class BoxUserCredentialPoliciesPrivateSchema(OneOfSchema):
    _DATA_TYPE = box_user_credential_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "box_user_credential_policies#1.0.0": box_user_credential_policies_1_0_0_private_schema,
        "box_user_credential_policies#1.1.0": box_user_credential_policies_1_1_0_private_schema,
    }
    type_field_remove = False
    connector_id = BOX_CONNECTOR_ID

    def get_obj_type(self, obj):
        return self._DATA_TYPE


box_user_credential_policies_schema = BoxUserCredentialPoliciesSchema(
    load_only=BoxUserCredentialPoliciesSchema.private
)
box_user_credential_policies_private_schema = BoxUserCredentialPoliciesPrivateSchema()


DataTypeSchema.register_schema(box_storage_policies_1_0_0_schema)
DataTypeSchema.register_schema(box_storage_policies_1_1_0_schema)
DataTypeSchema.register_schema(box_user_credential_policies_1_0_0_schema)
DataTypeSchema.register_schema(box_user_credential_policies_1_1_0_schema)
DataTypeSchema.register_schema(box_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=box_storage_policies_schema, private=box_storage_policies_private_schema
)

UserCredentialSchema.register_schema(
    public=box_user_credential_policies_schema,
    private=box_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(
    public=box_collection_policies_schema,
    private=box_collection_policies_private_schema,
)
