"""
Provides helpers for creating clients for the transfer and dns services

These functions must be called within an app context with valid Configuration
unless all optional arguments are given
"""
from typing import Optional

import globus_sdk
from globus_sdk.scopes import SearchScopes, TransferScopes
from globus_sdk.authorizers import GlobusAuthorizer

from globus.manager.api.models import Configuration
from globus.manager.api.utils.dns_client import DnsClient, DnsScopes
from globus.manager.appsync.client import AppsyncScopeBuilder, GraphQLClient


def get_auth_client(
    client_id: Optional[str] = None,
    secret: Optional[str] = None,
    environment: Optional[str] = None,
) -> globus_sdk.ConfidentialAppAuthClient:
    if not client_id:
        client_id = Configuration.query.filter_by(name="client_id").one().value
    assert client_id is not None
    if not secret:
        secret = Configuration.query.filter_by(name="client_secret").one().value
    assert secret is not None
    if not environment:
        environment = (
            Configuration.query.filter_by(name="environment").one().value.lower()
        )
    if environment == "default":
        environment = "production"

    return globus_sdk.ConfidentialAppAuthClient(
        client_id, secret, environment=environment
    )


def get_search_client(
    authorizer: Optional[GlobusAuthorizer] = None,
    client_id: Optional[str] = None,
    secret: Optional[str] = None,
    environment: Optional[str] = None,
) -> globus_sdk.SearchClient:
    if not environment:
        environment = (
            Configuration.query.filter_by(name="environment").one().value.lower()
        )
    if environment == "default":
        environment = "production"

    if not authorizer:
        authorizer = globus_sdk.ClientCredentialsAuthorizer(
            get_auth_client(
                client_id=client_id, secret=secret, environment=environment
            ),
            SearchScopes.globus_connect_server,
        )

    return globus_sdk.SearchClient(authorizer=authorizer, environment=environment)


def get_transfer_client(
    authorizer: Optional[GlobusAuthorizer] = None,
    client_id: Optional[str] = None,
    secret: Optional[str] = None,
    environment: Optional[str] = None,
) -> globus_sdk.TransferClient:
    if not environment:
        environment = (
            Configuration.query.filter_by(name="environment").one().value.lower()
        )

    if environment == "default":
        environment = "production"

    if not authorizer:
        authorizer = globus_sdk.ClientCredentialsAuthorizer(
            get_auth_client(
                client_id=client_id, secret=secret, environment=environment
            ),
            TransferScopes.all,
        )

    return globus_sdk.TransferClient(authorizer=authorizer, environment=environment)


def get_dns_client(
    authorizer: Optional[GlobusAuthorizer] = None,
    client_id: Optional[str] = None,
    secret: Optional[str] = None,
    environment: Optional[str] = None,
) -> DnsClient:
    if not environment:
        environment = (
            Configuration.query.filter_by(name="environment").one().value.lower()
        )

    if environment == "default":
        environment = "production"

    if not authorizer:
        authorizer = globus_sdk.ClientCredentialsAuthorizer(
            get_auth_client(
                client_id=client_id, secret=secret, environment=environment
            ),
            DnsScopes.manage_dns,
        )

    return DnsClient(authorizer=authorizer, environment=environment)


def get_appsync_client(
    client_id,
    authorizer: Optional[GlobusAuthorizer] = None,
    secret: Optional[str] = None,
    environment: Optional[str] = None,
) -> GraphQLClient:
    if not environment:
        environment = (
            Configuration.query.filter_by(name="environment").one().value.lower()
        )
    if environment == "default":
        environment = "production"

    assert environment is not None

    configuration = {
        "integration": {
            "client_id": "609c300a-3c94-4eae-ad04-ad30e2f5eea2",
            "url": "https://integration.gcs.globus.org",
        },
        "preview": {
            "client_id": "396f6867-18c9-45fa-8859-8519c1314d92",
            "url": "https://preview.gcs.globus.org",
        },
        "production": {
            "client_id": "41cffb35-201d-494d-8a18-f33261d4250f",
            "url": "https://gcs.globus.org",
        },
        "sandbox": {
            "client_id": "3f33d83f-ec0a-4190-887d-0622e7c4ee9a",
            "url": "https://sandbox.gcs.globus.org",
        },
        "staging": {
            "client_id": "9e1d2cae-2535-4cf8-b188-49da857e718d",
            "url": "https://staging.gcs.globus.org",
        },
        "test": {
            "client_id": "f5527eb3-127e-42ce-ad0e-a3328a783575",
            "url": "https://test.gcs.globus.org",
        },
    }

    appsync_scope_builder = AppsyncScopeBuilder(configuration[environment]["client_id"])
    scope = appsync_scope_builder.manager_configuration
    url = configuration[environment]["url"]

    if not authorizer:
        authorizer = globus_sdk.ClientCredentialsAuthorizer(
            get_auth_client(
                client_id=client_id, secret=secret, environment=environment
            ),
            scope,
        )

    return GraphQLClient(
        owner_id=client_id, environment=environment, authorizer=authorizer, base_url=url
    )
