"""
globus-connect-server endpoint migrate4
"""
import re
import os
import pwd
import json
import click
import textwrap
import marshmallow as ma
from flask import current_app
from typing import (
    Any,
    Callable,
    cast,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from typing_extensions import Protocol

import globus_sdk
import gcs_migration.check as check
from gcs_migration.util import do_link_auth_flow, get_api
import globus.manager.api.models as models
import globus.manager.api.schemas.api as schema
import gcs_migration.models as migration_models
from gcs_migration.models import MigrationPlan
from gcs_migration.commands.login import check_logged_in
from globus.manager.appsync.query import GraphQLError
from globus.manager.constants import CollectionType, RoleType
from globus.manager.api.blueprints.collections import allocate_domain
from globus.manager.api.app import make_app
from globus.manager.api.identity_mapping import (
    ExpressionIdentityMapping,
    ExternalIdentityMapping,
)
from globus.manager.api.identity_mapping.expression import (
    CompiledExpressionIdentityMappingDict,
)
from globus.manager.utils.keychain import NodeKey
from globus.manager.utils.path_restriction import PathRestrictions
from globus.manager.utils.obfuscate_int import obfuscate
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.api.blueprints.utils import (
    create_object,
    delete_object,
    patch_object,
    update_object,
)


class DomainAllocator:
    def __init__(self, collection_subdomain: str):
        self.domain_index = 0
        self.pre_allocated = 0
        self.collection_subdomain = collection_subdomain

    def preallocate_domains(self, endpoint: models.Endpoint, count: int) -> None:
        def update_index(ep):
            self.domain_index = ep.domain_index
            ep.domain_index += count
            self.pre_allocated = count

        update_object(endpoint, update_index)

    def allocate_domain(
        self,
        collection_type: Union[CollectionType, str],
        mapped_collection: Optional[models.Collection] = None,
    ) -> str:
        if self.pre_allocated > 0:
            ctype = CollectionType(collection_type).name
            self.domain_index += 1
            self.pre_allocated -= 1

            random_label = obfuscate(self.domain_index)
            if ctype == "guest":
                label = "g-" + random_label
            else:
                label = "m-" + random_label
            if ctype == "guest":
                if mapped_collection and mapped_collection.has_wildcard_domain:
                    return ".".join([label, mapped_collection.domain.domain_name])
            return ".".join([label, self.collection_subdomain])
        else:
            return allocate_domain(collection_type, mapped_collection)


class MigrationPlanDocs:
    """
    This class creates a derivative of gcs_migration.MigrationPlan for use
    throughout this command. It is the schema-validated version of the to_json()
    calls from each component of MigrationPlan, plus additional information that
    is needed for the migration.

    The class is structured very similarly to MigrationPlan with the exception
    that the schema-validated doc is stored at the 'doc' attribute of each component.
    So it looks like:

    MigrationPlanDocs:
    - storage_gateway:
      - doc: {...}
      mapped_collection:
      - v4_collection
        - id
        - model_exists
      - v5_collection
        - id
        - model_exists
      - doc: {...}
      - roles[]
        - v4_role_id
        - doc
      guest_collections[]:
      - v4_collection
        - id
        - model_exists
      - v5_collection
        - id
        - model_exists
      - doc: {...}
      - roles[]
        - v4_role_id
        - doc
      - acls[]
        - v4_acl_id
        - doc: {...}
      credentials[]:
        - identity_username
        - doc: {...}
    """

    class StorageGateway:
        def __init__(
            self, imported_gateway_plan: migration_models.StorageGateway
        ) -> None:
            self.id: Optional[str] = None
            self.doc = _load_doc(
                "storage gateway",
                imported_gateway_plan,
                schema.storage_gateway_schema,
            )

    class ModelInfo:
        def __init__(self, model_id: Optional[str] = None) -> None:
            self.id: Optional[str] = model_id
            self.exists: bool = False

    class Collection:
        def __init__(
            self,
            doc_name: str,
            imported_collection_plan: Union[
                migration_models.GuestCollection,
                migration_models.MappedCollection,
            ],
            role_docs: List["MigrationPlanDocs.Role"],
        ) -> None:
            self.v4_collection = MigrationPlanDocs.ModelInfo(
                imported_collection_plan.id
            )
            self.v5_collection = MigrationPlanDocs.ModelInfo()

            self.doc = _load_doc(
                doc_name,
                imported_collection_plan,
                schema.collection_schema,
            )

            self.roles = role_docs

    class MappedCollection(Collection):
        def __init__(
            self,
            imported_collection_plan: migration_models.MappedCollection,
            role_docs: List["MigrationPlanDocs.Role"],
        ) -> None:
            super().__init__("mapped collection", imported_collection_plan, role_docs)

    class GuestCollection(Collection):
        def __init__(
            self,
            imported_collection_plan: migration_models.GuestCollection,
            cred_doc: "MigrationPlanDocs.Credential",
            role_docs: List["MigrationPlanDocs.Role"],
            acl_docs: List["MigrationPlanDocs.ACL"],
        ) -> None:
            super().__init__("guest collection", imported_collection_plan, role_docs)
            self.credential = cred_doc
            self.acls = acl_docs

    class Role:
        def __init__(self, imported_role: migration_models.Role) -> None:
            self.v4_role_id = imported_role.id
            self.doc = _load_doc("role", imported_role, schema.role_schema)

    class ACL:
        def __init__(self, imported_acl: migration_models.ACL) -> None:
            self.v4_acl_id = imported_acl.id
            self.doc = _load_doc("ACL", imported_acl, None)

    class Credential:
        def __init__(self, imported_credential: migration_models.Credential) -> None:
            self.identity_username = imported_credential.identity_username
            # Inject a storage_gateway_id in order to pass the schema checks
            imported_credential.storage_gateway_id = (
                "DEADBEEF-DEAD-BEEF-DEAD-BEEFDEADBEAF"
            )
            self.doc = _load_doc(
                "credential", imported_credential, schema.user_credential_schema
            )

    def __init__(self, imported_plan: MigrationPlan) -> None:
        assert imported_plan.storage_gateway is not None

        self.storage_gateway = self.StorageGateway(imported_plan.storage_gateway)

        assert imported_plan.mapped_collection is not None
        role_docs = [
            self.Role(role) for role in imported_plan.mapped_collection.roles or []
        ]
        self.mapped_collection = self.MappedCollection(
            imported_plan.mapped_collection, role_docs
        )

        self.credentials = []
        for cred in imported_plan.credentials or []:
            # Credential
            self.credentials.append(self.Credential(cred))

        self.guest_collections = []
        for guest in imported_plan.guest_collections or []:
            # Credential
            assert guest.credential is not None
            cred_doc = self.Credential(guest.credential)
            # Roles
            role_docs = [self.Role(role) for role in guest.roles or []]
            # ACLs
            acls = guest.acls if guest.acls is not None else []
            acl_docs = [self.ACL(acl) for acl in acls]

            self.guest_collections.append(
                self.GuestCollection(guest, cred_doc, role_docs, acl_docs)
            )


def migrate4(verbose: bool, finalize: bool) -> None:
    # Run as root to get access to our log file.
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        click.get_current_context().exit(1)

    if not check_logged_in():
        do_link_auth_flow(
            session_params={"session_required_single_domain": "globusid.org"}
        )

    if verbose:
        os.environ["GCS_MANAGER_LOG_LEVEL"] = "DEBUG"

    app = make_app(app_config="globus.manager.api.config.endpoint_migrate4")
    models.db.init_app(app)

    app.logger.info("migrate4: start")

    with app.app_context():
        models.db.create_all()
        models.db.session.commit()

        for conf in models.Configuration.query:
            if conf.name == "node_key":
                value = NodeKey(**json.loads(conf.value))
            else:
                value = conf.value
            app.config[conf.name.upper()] = value

        if "CLIENT_ID" not in app.config:
            click.echo("This node has not been set up.", err=True)
            click.get_current_context().exit(1)

        # Load the keychain
        load_keychain()

        migrate4 = Migrate4()
        migrate4.run(finalize=finalize)


class Migrate4:
    CONNECTOR_MAP = {
        "7251f6c8-93c9-11eb-95ba-12704e0d6a4d": "ActiveScale",
        "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63": "Box",
        "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72": "Ceph",
        "56366b96-ac98-11e9-abac-9cb6d0d9fd63": "Google Cloud Storage",
        "976cf0cf-78c3-4aab-82d2-7c16adbcc281": "Google Drive",
        "e47b6920-ff57-11ea-8aaa-000c297ab3c2": "iRODS",
        "28ef55da-1f97-11eb-bdfd-12704e0d6a4d": "OneDrive",
        "145812c8-decc-41f1-83cf-bb2a85a2a70b": "POSIX",
        "052be037-7dda-4d20-b163-3077314dc3e6": "POSIX Staging",
        "7643e831-5f6c-4b47-a07f-8ee90f401d23": "S3",
        "7e3f3f5e-350c-4717-891a-2f451c24b0d4": "SpectraLogic BlackPearl",
    }
    SUPPORTED_CONNECTORS = ["POSIX", "S3", "SpectraLogic BlackPearl"]

    COLLECTION_DISPLAY_NAME_FORMAT = "Migration: {}"
    COLLECTION_KEYWORD_PREFIX = "migration_"
    COLLECTION_KEYWORD_FORMAT = COLLECTION_KEYWORD_PREFIX + "{}"

    def _has_finalize_been_run_before(self) -> bool:
        """
        Scan plan_docs to see if any collections have been finalized.
        """
        if self.plan_docs.mapped_collection.v4_collection.exists:
            return True
        guest_doc: MigrationPlanDocs.GuestCollection
        for guest_doc in self.plan_docs.guest_collections:
            if guest_doc.v4_collection.exists:
                return True
        return False

    def _retry_backoff(self, retry_context: globus_sdk.transport.RetryContext) -> float:
        """
        This is called by the SDK whenever the appsync client gets a retry-after response.
        Return the number of seconds that the SDK should wait before retrying.
        """
        wait_seconds: float = 30.0
        current_app.logger.debug(f"API calls throttled, waiting {wait_seconds}.")
        return wait_seconds

    def run(self, finalize: bool = False) -> None:
        # The WAF may throttle us during long migrations. If so, we want to keep retrying
        # until we are successful. 100 retries should be sufficient for more migrating more
        # than 1000 shares. Also, pause notifications so the DB doesn't change underneath us.
        with current_app.appsync_client.transport.tune(
            retry_backoff=self._retry_backoff, max_sleep=30, max_retries=100
        ), current_app.appsync_client.subscription_lock.exclusive():
            #
            # Phase 1: Load the plan and verify it.
            #
            self._load_migration_plan()
            self._check_subscription()
            self._verify_support_for_plan()
            self._verify_connector_properties()
            self._check_migration_status()

            #
            # Phase 2: Update models according to the migration plan. Skip this if
            # finalize has been run before.
            #
            if self._has_finalize_been_run_before() and not finalize:
                raise click.ClickException(
                    "This endpoint has already begun finalization. Please rerun with --finalize."
                )

            # Preallocate domains by bumping the domain index by the number of
            # collections, so that each collection creation doesn't require
            # an endpoint object update
            ep = models.Endpoint.query.one()
            self.domain_allocator = DomainAllocator(ep.collection_subdomain)
            self.domain_allocator.preallocate_domains(
                ep, 1 + len(self.plan_docs.guest_collections)
            )

            if not self._has_finalize_been_run_before():
                self._execute_migration_plan()
            else:
                click.echo("Finalization was previously run, skipping updates.")

            #
            # Phase 3: Delete extraneous guests on our temporary (v5) mapped collection so that finalize
            # doesn't break when try to delete it.
            #
            self._delete_extraneous_guests(
                cast(str, self.plan_docs.mapped_collection.v5_collection.id)
            )

            #
            # Phase 3: Finalize the migration, if requested
            #
            if finalize:
                transfer_client = get_api()
                self.finalize(transfer_client)

    def finalize(self, transfer_client: globus_sdk.TransferClient) -> None:
        """
        Finalize migration by transferring the v4 collection's id to the v5
        endpoint object in Transfer. This involves some low-level changes to
        the models to keep them coherent.
        """
        click.echo("Finalizing the migration.")

        # If we think the mapped collection has already been migrated, find its
        # model so we can use it as the mapped_collection when migrating the
        # other collections; otherwise we'll set it to None for now. In that
        # case, we'll set it to the model after migrating it.
        mapped_collection_model: Optional[models.Collection] = None
        if self.plan_docs.mapped_collection.v4_collection.exists:
            mapped_collection_model = models.Collection.query.filter(
                models.Collection.id
                == self.plan_docs.mapped_collection.v4_collection.id
            ).one()

        temporary_collection_model: models.Collection

        for collection in [
            cast(MigrationPlanDocs.Collection, self.plan_docs.mapped_collection)
        ] + cast(List[MigrationPlanDocs.Collection], self.plan_docs.guest_collections):
            # Skip collections that have already been migrated
            if collection.v4_collection.exists:
                click.echo(
                    f"Collection ID {collection.v4_collection.id} already migrated."
                )
                continue

            click.echo(
                f"Preparing to migrate collection ID {collection.v4_collection.id}."
            )

            # temporary_collection_model is the GCS model for the collection
            # before the uuid was migrated.
            temporary_collection_model = models.Collection.query.filter(
                models.Collection.id == collection.v5_collection.id
            ).one_or_none()

            if temporary_collection_model is None:
                raise click.ClickException(
                    f"Attempting to migrate collection {collection.v5_collection.id} "
                    "but it is not present in the database."
                )

            # Step 1: Set the gcs_private_temporary_migration_id value on each
            # of the v4 endpoint documents to allow the new v5 collection to
            # take ownership of its uuid. This is done with the token from the
            # owner of the v4 endpoint.

            try:
                # Set the flag in the transfer endpoint to allow this
                # endpoint's id to be transferred to the v5 collection's
                # endpoint
                transfer_client.update_endpoint(
                    cast(str, collection.v4_collection.id),
                    {"gcs_private_temporary_migration_id": collection.v5_collection.id},
                )
            except globus_sdk.exc.GlobusAPIError as e:
                # This could happen if the migration was already done for the
                # collection, but the models weren't committed to the DB.
                if e.http_status == 403:
                    # Verify that the error isn't because the collection owner is wrong
                    # in Transfer.
                    self._is_collection_owner_equal(
                        transfer_client,
                        cast(str, collection.v4_collection.id),
                        current_app.client_id,
                    )
                else:
                    raise click.ClickException(
                        f"Unable to migrate collection {collection.v4_collection.id}: {e}"
                    )

        for collection in [
            cast(MigrationPlanDocs.Collection, self.plan_docs.mapped_collection)
        ] + cast(List[MigrationPlanDocs.Collection], self.plan_docs.guest_collections):
            if collection.v4_collection.exists:
                click.echo(
                    f"Collection ID {collection.v4_collection.id} already migrated."
                )
                continue

            # temporary_collection_model is the GCS model for the collection
            # before the uuid was migrated.
            temporary_collection_model = models.Collection.query.filter(
                models.Collection.id == collection.v5_collection.id
            ).one_or_none()
            click.echo(f"Migrating collection ID {collection.v4_collection.id}.")
            # Step 2: Update the v5 collections using the private API to take
            # ownership of the v4 endpoint UUIDs. This returns a new collection
            # object and also destroys the v5 temporary collection in Transfer,
            # so we'll have an object that refers to stale data if we are terminated
            # before committing the changes to the database.
            try:
                current_app.transfer_client.post(
                    f"/private/endpoint/{collection.v5_collection.id}/migrate_id",
                    query_params={
                        "persistent_endpoint_id": collection.v4_collection.id
                    },
                )
            except Exception as e:
                current_app.logger.error(
                    f"Error migrating endpoint id {collection.v4_collection.id} "
                    f"to collection {collection.v5_collection.id}: {e}"
                )
                raise click.ClickException("Unable to migrate collection id")

            # Step 3: Add the newly migrated v4 collection model to appsync.

            domain_name = self.domain_allocator.allocate_domain(
                collection.doc["collection_type"]
            )
            temp_display_name = temporary_collection_model.display_name
            temp_prefix = self.COLLECTION_DISPLAY_NAME_FORMAT.format("")
            temporary_collection_model.display_name = temp_display_name.replace(
                temp_prefix, "", 1
            )

            # Collection.migrate_collection_uuid() will return a new
            # model that uses the persistent id
            final_model = temporary_collection_model.migrate_collection_uuid(
                cast(str, collection.v4_collection.id),
                domain_name,
                mapped_collection=mapped_collection_model,
            )

            try:
                # Add the new model to appsync
                final_model = current_app.appsync_client.create_object(final_model)
            except GraphQLError as e:
                if e.errors[0].get("errorType") == "CONFLICT":
                    if e.data is None:
                        raise click.ClickException(
                            f"Sync conflict for {collection.v4_collection.id} {e}"
                        )
                    # The model already exists in appsync from a previous
                    # migration attempt. Overwrite that with our model so we
                    # don't need to replace the transfer document and redo the
                    # scope text.
                    final_model.serial = e.data["items"][0]["serial"]
                    try:
                        final_model = current_app.appsync_client.update_object(
                            final_model
                        )
                    except GraphQLError as e:
                        raise click.ClickException(
                            f"Sync problem for {collection.v4_collection.id} {e}"
                        )

            models.db.session.add(final_model)
            models.db.session.commit()
            # Record that the v4 collection model exists
            collection.v4_collection.exists = True

            # If we migrated the mapped collection, track the model for
            # other guest migrations
            if final_model.is_mapped_collection:
                mapped_collection_model = final_model

        # We'll do the cleanup outside of the previous loop in order to handle recovery on
        # successive runs of migrate4, otherwise we'd miss cleaning up messes that happened
        # after the v4 model was committed.

        # Step 4: Update the roles on the v5 collection to be re-parented to the
        # collection with the persistent collection uuid.
        for collection in [
            cast(MigrationPlanDocs.Collection, self.plan_docs.mapped_collection)
        ] + cast(List[MigrationPlanDocs.Collection], self.plan_docs.guest_collections):
            # Skip collections if the tmp v5 collection has already been deleted
            if not collection.v5_collection.exists:
                click.echo(
                    f"Role reparenting already completed for collection {collection.v4_collection.id}."
                )
                continue
            click.echo(
                f"Reparenting roles for collection {collection.v5_collection.id}."
            )

            v4_collection_model = models.Collection.query.filter(
                ~models.Collection.is_deleted,  # type: ignore
                models.Collection.id == collection.v4_collection.id,
            ).one()

            for role in models.Role.query.filter(
                models.Role.resource == collection.v5_collection.id,
                ~models.Role.is_deleted,  # type: ignore
            ):
                # Move roles to the final model
                try:
                    update_object(
                        role, models.Role.migrate_collection_uuid, v4_collection_model
                    )
                except Exception as e:
                    current_app.logger.error(
                        "Could not update role object for collection "
                        f"{collection.v4_collection.id}: {e}"
                    )
                    raise click.ClickException(
                        f"Migration unable to update role {role.id} for collection "
                        f"{collection.v4_collection.id}"
                    )

        # Step 6: Delete the collection objects which use the temporary v5 UUIDs. Ordered
        # in reverse (guest before mapped so we can hard-delete things). We can not call
        # delete_object() on the temp collection model because it now shares resources with
        # the final v4 model.
        for plan in cast(
            List[MigrationPlanDocs.Collection], self.plan_docs.guest_collections
        ) + [cast(MigrationPlanDocs.Collection, self.plan_docs.mapped_collection)]:
            if not plan.v5_collection.exists:
                click.echo(
                    f"Temporary collection for {plan.v4_collection.id} already removed."
                )
                continue

            click.echo(f"Removing temporary collection {plan.v5_collection.id}.")
            model: models.Collection = models.Collection.query.filter(
                models.Collection.id == plan.v5_collection.id
            ).one()
            try:
                self._delete_collection(model, migrated=True)
            except Exception as e:
                raise click.ClickException(
                    f"Unable to delete temporary collection {model.id}: {e}"
                )
            plan.v5_collection.exists = False

        # Step 7: Remove migration keywords from final v4 collection models. We do it here instead
        # of when we created the v4 collection model in order to handle endpoints finalized with
        # older versions of migrate4.
        for collection in [
            cast(
                MigrationPlanDocs.Collection,
                self.plan_docs.mapped_collection,
            )
        ] + cast(
            List[
                MigrationPlanDocs.Collection,
            ],
            self.plan_docs.guest_collections,
        ):
            final_model = models.Collection.query.filter(
                models.Collection.id == collection.v4_collection.id
            ).one()

            click.echo(
                f"Removing migration keywords on collection {collection.v4_collection.id}."
            )

            update_doc = {
                "DATA_TYPE": "collection#1.4.0",
                "keywords": collection.doc.get("keywords", []),
            }
            patch_object(final_model, update_doc)

    def _is_collection_owner_equal(
        self,
        transfer_client: globus_sdk.TransferClient,
        collection_id: str,
        owner_id: str,
    ) -> bool:
        try:
            res = transfer_client.get_endpoint(collection_id)
            if res["owner_id"] == owner_id:
                current_app.logger.info(
                    f"Collection {collection_id} is already "
                    "migrated in the Globus Transfer"
                )
                return True
            else:
                current_app.logger.error(
                    f"Collection {collection_id} can not be migrated, it is "
                    f"owned by globus user {res['owner_id']}"
                )
                raise click.ClickException(
                    "Unable to migrate collection {collection_id}"
                )

        except Exception as e:
            current_app.logger.info(
                f"Unable to to prepare endpoint {collection_id} for migration. "
                f"The endpoint is not accessible in the transfer service: {e}"
            )
            raise click.ClickException(f"Unable to migrate collection {collection_id}")

    def _load_migration_plan(self) -> None:
        """
        Load the migration plan from disk and verify it.

        The important instance variables setup here and used in the remainder
        of this class are:

        - self.imported_plan is a gcs_migration.MigrationPlan
        - self.plan_docs is an instance of MigrationPlanDocs
        """
        click.echo("Loading the migration plan from disk.")

        # Load from disk
        self.imported_plan = _import_migration_plan()
        # Run internal gcs_migration checks
        _check_imported_plan(self.imported_plan)
        # No support for cross-environment migration
        self._check_globus_environment(
            current_app.environment, self.imported_plan.environment
        )
        # Create and verify the component docs for the endpoint sync
        self.plan_docs = MigrationPlanDocs(self.imported_plan)

    def _check_subscription(self) -> None:
        """
        Check for a subscription to cover paid features
        """
        click.echo("Checking subscription")
        sub_id = current_app.subscription

        # We assume any subscription covers all features
        if sub_id is not None:
            return

        # Connector ID
        connector_id = self.plan_docs.storage_gateway.doc["connector_id"]
        if self.CONNECTOR_MAP.get(connector_id, None) != "POSIX":
            click.echo(
                "The migration plan includes a connector that requires a subscription. "
                "Set a subscription on this endpoint and rerun this command.",
                err=True,
            )
            click.get_current_context().exit(0)

        # Guest collections
        if self.plan_docs.guest_collections:
            click.echo(
                "The migration plan includes guest collections which requires a subscription. "
                "Set a subscription on this endpoint and rerun this command.",
                err=True,
            )
            click.get_current_context().exit(0)

        collection = self.plan_docs.mapped_collection
        if len(collection.roles) > 1:
            click.echo(
                "The migration plan includes roles which require a subscription. "
                "Set a subscription on this endpoint and rerun this command.",
                err=True,
            )
            click.get_current_context().exit(0)

        if collection.doc.get("enable_https"):
            click.echo(
                "The migration plan includes the https interface which requires a subscription. "
                "Set a subscription on this endpoint and rerun this command.",
                err=True,
            )
            click.get_current_context().exit(0)

    def _verify_support_for_plan(self) -> None:
        """
        Check the migration plan doesn't ask for anything we don't yet support.
        """
        click.echo("Checking the migration plan for unsupported features.")

        storage_gateway = self.plan_docs.storage_gateway
        # No support for HA gateways
        if storage_gateway.doc.get("high_assurance", False) is True:
            click.echo(
                "High assurance gateways are not supported with migration.", err=True
            )
            click.get_current_context().exit(1)

        # Make sure it is a known connector_id
        connector_id = storage_gateway.doc["connector_id"]
        connector_type = self.CONNECTOR_MAP.get(connector_id)

        if connector_type is None:
            click.echo(
                f"Unknown connector id {storage_gateway.doc['connector_id']}.", err=True
            )
            click.get_current_context().exit(1)

        # Make sure it is a supported connector_id
        if connector_type not in self.SUPPORTED_CONNECTORS:
            click.echo(
                f"{connector_type} connector is not yet supported for migration.",
                err=True,
            )
            click.get_current_context().exit(1)

    def _verify_posix_connector_properties(self) -> None:
        """
        Verify anything about the migration plan that is specific to POSIX.

        We need to verify that usernames in guest collection user credentials are
        valid local usernames.
        """
        click.echo("Verifying properties specific to the POSIX connector.")

        local_usernames = []
        for guest in self.plan_docs.guest_collections or []:
            if guest.credential.doc["username"] not in local_usernames:
                local_usernames.append(guest.credential.doc["username"])

        missing_usernames = []
        for username in local_usernames:
            try:
                pwd.getpwnam(username)
            except KeyError:
                missing_usernames.append(username)

        if missing_usernames:
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "These POSIX usernames are not available on this system. Please "
                        "add these accounts to this node and rerun this command."
                    )
                    + [""]
                    + textwrap.wrap(f"Missing accounts: {', '.join(missing_usernames)}")
                ),
                err=True,
            )
            click.get_current_context().exit(1)

    def _verify_connector_properties(self) -> None:
        switcher = {
            "145812c8-decc-41f1-83cf-bb2a85a2a70b": self._verify_posix_connector_properties,
        }

        func = switcher.get(self.plan_docs.storage_gateway.doc["connector_id"])
        if func:
            func()

    def _lookup_collection_plan(
        self, collection_id: str
    ) -> Optional[MigrationPlanDocs.Collection]:
        """
        Given a collection ID, determine if it is part of the migration plan.

        The collection may match in either of these cases:
        - The ID matches a collection doc's v4 collection ID. This is a final model.
        - The ID matches a collection doc's v5 collection ID. This is a temp model.

        The caller can determine if this is a temporary or final model by checking if the
        IDs match.
        """
        for doc in [
            cast(MigrationPlanDocs.Collection, self.plan_docs.mapped_collection)
        ] + cast(List[MigrationPlanDocs.Collection], self.plan_docs.guest_collections):
            # Check if this is a final (migrated) collection model.
            if collection_id == doc.v4_collection.id:
                return doc
            # Check if this is a temp (unmigrated) collection model.
            if collection_id == doc.v5_collection.id:
                return doc
        return None

    def _check_migration_status(self) -> None:
        """
        Search the current endpoint for the gateway and collections described in the migration plan.

        This will fill in self.plan_docs with the gateway and collection IDs and whether the temp and
        final models of each collection exists.
        """
        click.echo("Checking the status of the migration.")

        # Since the v4 endpoint display name and the gateway display name can change, we'll look for
        # collections in our plans and use their gateway value to locate the gateway.
        model: models.Collection
        for model in models.Collection.query.filter(
            ~models.Collection.is_deleted,  # type: ignore
        ):
            # This will hit only on final (v4) collection models because we haven't filled
            # in temporary (v5) collection model IDs yet.
            plan = self._lookup_collection_plan(model.id)
            if plan:
                assert plan.v4_collection.id == model.id
                assert not plan.v4_collection.exists
                plan.v4_collection.exists = True

                # Save off the storage gateway ID
                if self.plan_docs.storage_gateway.id:
                    assert self.plan_docs.storage_gateway.id == model.storage_gateway_id
                else:
                    self.plan_docs.storage_gateway.id = cast(
                        str, model.storage_gateway_id
                    )
                continue

            # Now check if this model is a temporary (v5) collection in the plan. The migration
            # keywords should exist on temporary collection models. It may exist on final collection
            # models too but we handled that match above.
            keyword_id = self._get_id_from_migration_keywords(
                getattr(model, "keywords", [])
            )

            # Collections on different gateways or guests created on our mapped collection,
            # which we'll delete later since we won't record them in the plan.
            if not keyword_id:
                continue

            plan = self._lookup_collection_plan(keyword_id)

            # Collections removed from our plan or that exist on another gateway in another migration plan.
            if not plan:
                continue

            # It is possible to have multiple collections with the same migration keywords, for
            # example, if the user credential changes between successive runs of migrate4. Make sure
            # this collection matches before we record it.
            if model.collection_type == CollectionType.guest.value:
                guest_plan = cast(MigrationPlanDocs.GuestCollection, plan)
                user_credential = cast(models.UserCredential, model.user_credential)
                if user_credential.identity_id != str(
                    guest_plan.credential.doc["identity_id"]
                ):
                    continue
                if user_credential.username != str(
                    guest_plan.credential.doc["username"]
                ):
                    continue

            # We have a match! This is a temporary (v5) collection model in our plan.
            assert not plan.v5_collection.exists
            plan.v5_collection.exists = True
            plan.v5_collection.id = model.id

            # Save off the storage gateway ID
            if self.plan_docs.storage_gateway.id:
                assert self.plan_docs.storage_gateway.id == model.storage_gateway_id
            else:
                self.plan_docs.storage_gateway.id = cast(str, model.storage_gateway_id)
            continue

        # If we didn't find any existing collections from our plans, we'll need to find the gateway by
        # display name.
        if not self.plan_docs.storage_gateway.id:
            gateway_model: models.StorageGateway = models.StorageGateway.query.filter(
                ~models.StorageGateway.deleted,  # type: ignore
                models.StorageGateway.display_name
                == self.plan_docs.storage_gateway.doc["display_name"],
            ).one_or_none()

            if gateway_model:
                self.plan_docs.storage_gateway.id = gateway_model.storage_gateway_id

    def _execute_migration_plan(self) -> None:
        click.echo("Executing the migration plan.")

        gateway_model = self._sync_storage_gateway(self.plan_docs.storage_gateway)
        mapped_collection_model = self._sync_mapped_collection(
            self.plan_docs.mapped_collection, gateway_model
        )
        self._sync_user_credentials(self.plan_docs.credentials, gateway_model)
        for guest_plan in self.plan_docs.guest_collections:
            self._sync_guest_collection(
                guest_plan, gateway_model, mapped_collection_model
            )

    def _delete_extraneous_guests(self, mapped_collection_id: str) -> None:
        """
        Delete guest collections on our mapped collection that are not in our migration plan.

        We need to make sure there are no guest collections, neither not-deleted nor
        soft-deleted, on our mapped collection so that we can hard-delete it later.
        """
        for guest_model in models.Collection.query.filter(
            models.Collection.collection_type == CollectionType.guest,
            models.Collection.mapped_collection_id == mapped_collection_id,
        ):
            # Skip guests in our plan unless they are soft-deleted
            if not guest_model.deleted:
                plan = self._lookup_collection_plan(guest_model.id)
                if plan:
                    continue

            msg = "soft-deleted" if guest_model.deleted else "extraneous"
            click.echo(f"Deleting {msg} v5 guest collection {guest_model.id}")

            try:
                self._delete_collection(guest_model)
            except Exception as e:
                raise click.ClickException(
                    f"Unable to delete guest collection {guest_model.id}: {e}"
                )

    def _sync_storage_gateway(
        self, plan: MigrationPlanDocs.StorageGateway
    ) -> models.StorageGateway:
        """
        Create or update the storage gateway

        We name the storage gateway according to the migration plan, without any special
        formatting since it is not a public object and there is no corresponding v4 version.

        We expect this display name to remain unchange so that we can look it up on successive
        iterations of this command. If the gateway display_name is changed, we may assume it
        does not exist and we recreate it (and everything below it).
        """
        click.echo("Syncing the storage gateway")

        if not plan.id:
            click.echo(".Creating the new v5 storage gateway")
            gateway_model = self._create_storage_gateway(plan.doc)
            plan.id = gateway_model.id
            return gateway_model

        gateway_model = models.StorageGateway.query.filter(
            ~models.StorageGateway.deleted,  # type: ignore
            models.StorageGateway.id == plan.id,
        ).one()

        update_doc = self._create_gateway_update_doc(plan.doc, gateway_model)
        if update_doc:
            click.echo(".Updating the existing v5 storage gateway")
            self._update_storage_gateway(update_doc, gateway_model)
        return gateway_model

    def _create_storage_gateway(
        self,
        doc: Dict[str, Any],
    ) -> models.StorageGateway:
        current_app.logger.debug(f"Creating storage gateway: {doc}")
        try:
            cls = models.StorageGateway.polymorphic_class(doc["connector_id"])
            gateway = cls(doc=doc)
        except (TypeError, ValueError) as e:
            click.echo(
                "Error creating storage gateway model: {}".format(str(e)), err=True
            )
            click.get_current_context().exit(1)
        create_object(gateway)
        return gateway

    def _update_storage_gateway(
        self, update_doc: Dict[str, Any], model: models.StorageGateway
    ) -> None:
        current_app.logger.debug(f"Updating storage gateway: {update_doc}")
        patch_object(model, update_doc)

    def _create_gateway_update_doc(
        self, doc: Dict[str, Any], model: models.StorageGateway
    ) -> Dict[str, Any]:
        """
        Create a doc to update the storage gateway based on differences between the
        model and doc. If no changes are detcted, we return {}. Limiting updates
        significantly speeds up succesive runs of migrate4.

        We only check the values that gcs_migration's gateway model provides.
        """
        update_doc: Dict[str, Any] = {}
        for attr in [
            "allowed_domains",
            "display_name",
            "users_allow",
            "users_deny",
        ]:
            org_value = getattr(model, attr, None)
            new_value = doc.get(attr)
            if (org_value is None) ^ (new_value is None) or (
                org_value is not None
                and new_value is not None
                and set(org_value) != set(new_value)
            ):
                update_doc[attr] = new_value

        model_value: Optional[
            List[Union[ExpressionIdentityMapping, ExternalIdentityMapping]]
        ] = getattr(model, "identity_mappings", None)
        doc_value = doc.get("identity_mappings")
        if self._cmp_identity_mappings(model_value, doc_value) != 0:
            update_doc["identity_mappings"] = doc_value

        model_value = getattr(model, "restrict_paths", None)
        doc_value = doc.get("restrict_paths")
        if self._cmp_restrict_paths(model_value, doc_value) != 0:
            update_doc["restrict_paths"] = doc_value

        for attr in ["groups_allow", "groups_deny"]:
            org_value = getattr(model, attr, None)
            new_value = doc["policies"].get(attr)
            if (org_value is None) ^ (new_value is None) or (
                org_value is not None
                and new_value is not None
                and set(org_value) != set(new_value)
            ):
                if "policies" not in update_doc:
                    update_doc["policies"] = {}
                update_doc["policies"][attr] = new_value

        # If there are changes, fill out the boilerplate
        if update_doc:
            update_doc["DATA_TYPE"] = "storage_gateway#1.1.0"
            if update_doc.get("policies") is None:
                update_doc["policies"] = {}
            if "DATA_TYPE" not in update_doc["policies"]:
                update_doc["policies"]["DATA_TYPE"] = model.POLICY_DATA_TYPE

        return update_doc

    def _sync_mapped_collection(
        self,
        plan: MigrationPlanDocs.MappedCollection,
        storage_gateway: models.StorageGateway,
    ) -> models.Collection:
        click.echo(
            f"Syncing v4 endpoint {plan.v4_collection.id} to v5 mapped collection"
        )

        # _sync_collection() will create the collection and the roles
        return self._sync_collection(
            plan,
            storage_gateway,
            authorized_identity={"id": plan.doc["identity_id"]},
        )

    def _sync_guest_collection(
        self,
        plan: MigrationPlanDocs.GuestCollection,
        storage_gateway: models.StorageGateway,
        mapped_collection: models.Collection,
    ) -> None:
        click.echo(f"Syncing v4 share {plan.v4_collection.id} to v5 guest collection")
        # Get this collection's user credential
        user_credental = self._sync_user_credential(plan.credential, storage_gateway)

        # _sync_collection() will create the collection and the roles
        collection = self._sync_collection(
            plan,
            storage_gateway,
            authorized_identity={
                "id": plan.doc["identity_id"],
                "username": plan.credential.identity_username,
                "status": "used",
            },
            mapped_collection=mapped_collection,
            user_credential=user_credental,
        )

        # Sync this collection's ACLs
        self._sync_acls(plan.acls, collection)
        return

    def _sync_user_credentials(self, plans, storage_gateway):
        # Create user credentials
        for plan in plans:
            self._sync_user_credential(plan, storage_gateway)

    def _sync_collection(
        self,
        plan: MigrationPlanDocs.Collection,
        storage_gateway: models.StorageGateway,
        authorized_identity: Dict[str, Any],
        mapped_collection: Optional[models.Collection] = None,
        user_credential: Optional[models.UserCredential] = None,
    ) -> models.StorageGateway:
        """
        Create or update the collection

        We create the collection with a temporary ID. Later when the admin is ready, we'll
        provide the ability to switch the collection ID to the ID listed in the migration plan.

        We'll format the collection display name in order to avoid conflicts with the production
        v4 instance. We add a special keyword to the collection so that we can find the collection
        on successive iterations of this command.
        """
        if not plan.v5_collection.exists:
            click.echo(".Creating the new v5 collection")
            model = self._create_collection(
                plan.doc,
                cast(str, plan.v4_collection.id),
                storage_gateway,
                authorized_identity,
                mapped_collection,
                user_credential,
            )
            plan.v5_collection.id = model.id
            plan.v5_collection.exists = True
        else:
            model = models.Collection.query.filter(
                ~models.Collection.is_deleted,  # type: ignore
                models.Collection.id == plan.v5_collection.id,
            ).one()

            # Update the v5 collection model
            update_doc = self._create_collection_update_doc(
                plan.doc, cast(str, plan.v4_collection.id), model
            )
            if update_doc:
                click.echo(".Updating the v5 collection")
                self._update_collection(update_doc, model)

        # Sync this collection's roles
        self._sync_roles(plan.roles, model)
        return model

    def _create_collection(
        self,
        doc: Dict[str, Any],
        v4_collection_id: str,
        storage_gateway: models.StorageGateway,
        authorized_identity: Dict[str, Any],
        mapped_collection: Optional[models.Collection],
        user_credential: Optional[models.UserCredential],
    ) -> models.Collection:
        current_app.logger.debug(f"Creating collection: {doc}")

        modified_display_name = self._migration_display_name(doc["display_name"])
        modified_keywords = self._insert_migration_keyword(
            doc.get("keywords", []), v4_collection_id
        )

        create_doc = {
            **doc,
            "display_name": modified_display_name,
            "keywords": modified_keywords,
            "collection_type": doc["collection_type"],
        }
        if mapped_collection:
            create_doc["mapped_collection_id"] = mapped_collection.id

        try:
            domain_name = self.domain_allocator.allocate_domain(
                create_doc["collection_type"]
            )
            collection = models.Collection.from_collections_post(
                storage_gateway=storage_gateway,
                mapped_collection=mapped_collection,
                endpoint=models.Endpoint.query.one(),
                user_credential=user_credential,
                domain_name=domain_name,
                authorized_identity=authorized_identity,
                doc=create_doc,
            )
        except Exception as e:
            click.echo(
                f"Error creating collection: {str(e)}",
                err=True,
            )
            click.get_current_context().exit(1)
        create_object(collection)
        # If the endpoint is managed, there will be an owner role associated with
        # the collection. If we don't commit it here, it'll be in the DB but not
        # in appsync. migrate4 won't try to fix it later because it'll see it in
        # the DB.
        if collection.owner_role:
            create_object(collection.owner_role)
        return collection

    def _update_collection(
        self, update_doc: Dict[str, Any], model: models.Collection
    ) -> None:
        current_app.logger.debug(f"Updating collection {update_doc}")
        patch_object(model, update_doc)

    def _create_collection_update_doc(
        self,
        doc: Dict[str, Any],
        v4_collection_id: str,
        model: models.Collection,
    ) -> Dict[str, Any]:
        """
        Create a doc to update the collection based on differences between the model and doc.
        If no changes are detected, we return {}. Limiting updates significantly speeds up
        succesive runs of migrate4.

        We only check the values that gcs_migration's collection model provides.
        """
        update_doc: Dict[str, Any] = {}

        #
        # Common endpoint attributes
        #

        # display_name comparison
        modified_display_name = self._migration_display_name(doc["display_name"])
        if model.display_name != modified_display_name:
            update_doc["display_name"] = modified_display_name

        for attr in [
            "description",
            "info_link",
            "contact_info",
            "contact_email",
            "organization",
            "department",
        ]:
            if getattr(model, attr, None) != doc.get(attr):
                update_doc[attr] = doc.get(attr)

        # keywords comparison
        modified_keywords = self._insert_migration_keyword(
            doc.get("keywords", []), v4_collection_id
        )
        if set(getattr(model, "keywords", [])) != set(modified_keywords):
            update_doc["keywords"] = modified_keywords

        #
        # Mapped collection attributes
        #

        # Skipping these attributes:
        # - collection_base_path: immutable
        # - identity_id: immutable
        if model.collection_type == CollectionType.mapped.value:
            for attr, default in [
                ("default_directory", None),
                ("public", None),
                ("force_encryption", None),
                ("force_verify", None),
                ("disable_verify", None),
                ("allow_guest_collections", None),
                ("enable_https", True),
                ("user_message", None),
                ("user_message_link", None),
            ]:
                if getattr(model, attr, None) != doc.get(attr, default):
                    update_doc[attr] = doc.get(attr, default)

            for attr in ["sharing_users_allow", "sharing_users_deny"]:
                org_value = getattr(model, attr, None)
                new_value = doc.get(attr)
                if (org_value is None) ^ (new_value is None) or (
                    org_value is not None
                    and new_value is not None
                    and set(org_value) != set(new_value)
                ):
                    update_doc[attr] = new_value

            for attr in ["sharing_groups_allow", "sharing_groups_deny"]:
                org_value = getattr(model, attr, None)
                new_value = doc["policies"].get(attr)
                if (org_value is None) ^ (new_value is None) or (
                    org_value is not None
                    and new_value is not None
                    and set(org_value) != set(new_value)
                ):
                    if "policies" not in update_doc:
                        update_doc["policies"] = {}
                    update_doc["policies"][attr] = new_value

            model_value = model.sharing_restrict_paths
            doc_value = doc.get("sharing_restrict_paths")
            if self._cmp_sharing_restrict_paths(model_value, doc_value) != 0:
                update_doc["sharing_restrict_paths"] = doc.get("sharing_restrict_paths")

            inherited_policies = [
                "user_message",
                "user_message_link",
                "force_encryption",
                "force_verify",
                "disable_verify",
                "enable_https",
            ]
            for ip in inherited_policies:
                if ip in update_doc:
                    self.update_guests = True
                    break

        #
        # Guest collection attributes
        #

        # Skipping these attributes:
        # - collection_base_path: immutable
        # - identity_id: immutable
        # - mapped_collection_id: immutable
        if model.collection_type == CollectionType.guest.value:
            for attr, default in [
                ("default_directory", None),
                ("public", None),
                ("force_encryption", False),
                ("force_verify", False),
                ("disable_verify", False),
                ("enable_https", True),
            ]:
                if getattr(self, "update_guests", False) or getattr(
                    model, attr, None
                ) != doc.get(attr, default):
                    update_doc[attr] = doc.get(attr, default)

        if update_doc:
            update_doc["DATA_TYPE"] = "collection#1.4.0"
            if "policies" not in update_doc:
                update_doc["policies"] = {}
            if "DATA_TYPE" not in update_doc["policies"]:
                update_doc["policies"]["DATA_TYPE"] = model.POLICY_DATA_TYPE
        return update_doc

    def _delete_collection(
        self, model: models.Collection, migrated: bool = False
    ) -> None:
        """
        Safely hard delete the collection associated with 'model'. The collection may exist,
        be soft-deleted or have a pending_update delete.

        :param bool migrated:
            True if model is a temporary collection model and the final model has already
            been created, False otherwise.
        """
        # If the collection exists or is pending delete, it may still have roles.
        if not model.deleted:
            for r in models.Role.query.filter(models.Role.resource == model.id):
                if r.primary_role:
                    r.primary_role = False
                    r.refresh_configuration_hash()
                # (Hard) Delete the role
                delete_object(r, collection_deleted=True)

        # (Hard) Delete the collection
        delete_object(model, delete_transfer_endpoint=not migrated)

    def _delete_soft_deleted_collection(self, model: models.Collection) -> None:
        """
        Hard delete the soft-deleted collection associated with 'model'.
        """
        # The model must be soft-deleted. It can not be a pending delete.
        assert model.deleted
        # soft-deleted collection have no roles
        assert not models.Role.query.filter(models.Role.resource == model.id).all()
        # (Hard) Delete the collection
        delete_object(model, update=False)

    def _sync_user_credential(
        self, plan: MigrationPlanDocs.Credential, storage_gateway: models.StorageGateway
    ) -> models.UserCredential:
        """
        Sync the user credentials
        """
        click.echo(".Syncing the user credential")

        # Check if the user credential already exists
        current_app.logger.debug(f"Checking if user credential exists: {plan.doc}")

        cred = models.UserCredential.query.filter(
            ~models.UserCredential.is_deleted,  # type: ignore
            models.UserCredential.storage_gateway_id == storage_gateway.id,
            models.UserCredential.identity_id == str(plan.doc["identity_id"]),
            models.UserCredential.username == plan.doc["username"],
        ).one_or_none()

        if cred:
            return cred

        current_app.logger.debug("Creating user credential")
        click.echo("..Creating the new user credential")
        try:
            cls = models.UserCredential.polymorphic_class(str(plan.doc["connector_id"]))
            doc = {
                **plan.doc,
                "storage_gateway_id": str(storage_gateway.id),
                "identity_id": str(plan.doc["identity_id"]),
                "connector_id": str(plan.doc["connector_id"]),
            }
            cred = cls(doc=doc, storage_gateway=storage_gateway)
        except (TypeError, ValueError) as e:
            click.echo(
                "Error creating user credential model: {}".format(str(e)), err=True
            )
            click.get_current_context().exit(1)
        create_object(cred)
        return cred

    def _sync_acls(
        self, acl_plans: List[MigrationPlanDocs.ACL], collection: models.Collection
    ) -> None:
        """
        For each explicit ACL on the v4 share, provide a matching ACL on the v5 guest collection
        and remove any ACLs from the v5 guest collection that were not on the v4 share.

        As of gcs-migration 0.18 and gcs 5.4.47, role_id and role_type are always set to 'None' in
        the ACL plans, however, 'id' is set to 'None' when the v4 ACL is an implicit ACL. We aim to
        match replicate ACLs; we ignore ACLs provided implicitly by roles.
        """
        click.echo(f".Syncing ACLs on v5 guest collection {collection.id}")

        # Remove ACLs associated with roles
        needed_acls = [acl for acl in acl_plans if acl.v4_acl_id is not None]

        # Remove ACLs associated with roles
        existing_acls = [
            acl for acl in self._lookup_all_acls(collection.id) if acl["id"] is not None
        ]

        # Phase I: Create missing ACLs
        for needed_acl in needed_acls:
            matching_acl = self._lookup_acl(
                needed_acl.doc["principal"],
                needed_acl.doc["principal_type"],
                needed_acl.doc["path"],
                existing_acls,
            )
            if not matching_acl:
                click.echo(f"..Creating v5 ACL for v4 ACL {needed_acl.v4_acl_id}")
                self._create_acl(needed_acl.doc, collection.id)
            elif needed_acl.doc["permissions"] != matching_acl["permissions"]:
                click.echo(f"..Updating v5 ACL {matching_acl['id']}")
                self._update_acl(needed_acl.doc, matching_acl, collection.id)

        # Phase II: Delete extraneous ACLs
        for existing_acl in existing_acls:
            if not self._is_acl_in_plans(needed_acls, existing_acl):
                click.echo(f"..Deleting extraneous v5 ACL {existing_acl['id']}")
                self._delete_acl(existing_acl["id"], collection.id)

    def _lookup_acl(
        self,
        principal: Dict[str, Any],
        principal_type: str,
        path: str,
        existing_acls: List[Dict[str, Any]],
    ) -> Optional[Dict[str, Any]]:
        """
        Return the first matching ACL.
        """
        for acl in existing_acls:
            if (
                acl["principal"] == principal
                and acl["principal_type"] == principal_type
                and acl["path"] == path
            ):
                return acl
        return None

    def _lookup_all_acls(self, collection_id: str) -> List[Dict[str, Any]]:
        """
        Fetch all ACLs for the collection
        """
        try:
            transfer_client = current_app.transfer_client
            res = transfer_client.endpoint_acl_list(collection_id)
        except globus_sdk.TransferAPIError as e:
            click.echo(
                f"Could not get ACLs for guest collection {collection_id}: {e}",
                err=True,
            )
            click.get_current_context().exit(1)
        return res["DATA"]

    def _is_acl_in_plans(
        self, acl_plans: List[MigrationPlanDocs.ACL], acl: Dict[str, Any]
    ) -> bool:
        for plan in acl_plans:
            if (
                acl["principal"] == plan.doc["principal"]
                and acl["principal_type"] == plan.doc["principal_type"]
                and acl["path"] == plan.doc["path"]
            ):
                return True
        return False

    def _create_acl(self, doc: Dict[str, Any], collection_id: str) -> None:
        try:
            transfer_client = current_app.transfer_client
            transfer_client.add_endpoint_acl_rule(collection_id, doc)
        except globus_sdk.TransferAPIError as e:
            click.echo(
                f"Could not add ACL to guest collection {collection_id}: {e}",
                err=True,
            )
            click.get_current_context().exit(1)

    def _update_acl(
        self, doc: Dict[str, Any], acl: Dict[str, Any], collection_id: str
    ) -> None:
        try:
            transfer_client = current_app.transfer_client
            transfer_client.update_endpoint_acl_rule(collection_id, acl["id"], doc)
        except globus_sdk.TransferAPIError as e:
            click.echo(
                f"Could not update ACL on guest collection {collection_id}: {e}",
                err=True,
            )
            click.get_current_context().exit(1)

    def _delete_acl(self, acl_id, collection_id) -> None:
        try:
            transfer_client = current_app.transfer_client
            transfer_client.delete_endpoint_acl_rule(collection_id, acl_id)
        except globus_sdk.TransferAPIError as e:
            click.echo(
                f"Could not delete ACL {acl_id} on guest collection {collection_id}: {e}",
                err=True,
            )
            click.get_current_context().exit(1)

    def _sync_roles(
        self, role_plans: List[MigrationPlanDocs.Role], collection: models.Collection
    ) -> None:
        click.echo(
            f".Syncing roles on v5 {collection.collection_type} collection {collection.id}"
        )

        all_role_ids: List[str] = []

        # We can not update roles, just delete them
        # Phase I: Delete extraneous roles
        for role in self._lookup_all_roles(collection.id):
            if not self._is_role_in_plans(role_plans, role) and not role.primary_role:
                click.echo(f"..Deleting extraneous v5 role {role.id}")
                self._delete_role(role)
            else:
                all_role_ids.append(role.id)

        def is_implicit_role(
            collection: models.Collection, role: Dict[str, Any]
        ) -> bool:
            return (
                role["principal"] in [collection.identity_id, current_app.client_id]
                and role["principal_type"] == "identity"
                and role["role"] == "administrator"
            )

        # Phase II: Delete roles in transfer that don't match any in GCS db
        for role in current_app.transfer_client.endpoint_role_list(
            collection.id
        ).data.get("DATA", []):
            role_id = role.get("id")
            if role_id not in all_role_ids and not is_implicit_role(collection, role):
                click.echo(f"..Deleting unknown v5 role {role_id}")
                current_app.transfer_client.delete_endpoint_role(collection.id, role_id)

        # Phase III: Create missing roles
        for plan in role_plans:
            old_role = self._lookup_role(
                plan.doc["principal"], plan.doc["role"], collection.id
            )
            if not old_role:
                click.echo(f"..Creating v5 role for v4 role {plan.v4_role_id}")
                self._create_role(plan.doc, collection.id)

    def _lookup_role(
        self, principal: Dict[str, Any], role: str, collection_id: str
    ) -> Optional[models.Role]:
        """
        Lookup a single role in the database.
        """
        return models.Role.query.filter(
            ~models.Role.is_deleted,  # type: ignore
            models.Role.principal == principal,
            models.Role.role_name == role,
            models.Role.resource.is_(collection_id),  # type: ignore
        ).one_or_none()

    def _lookup_all_roles(self, collection_id: str) -> Iterator[models.Role]:
        """
        Lookup all roles for a collection in the database.
        """
        return models.Role.query.filter(
            ~models.Role.is_deleted,  # type: ignore
            models.Role.resource.is_(collection_id),  # type: ignore
        ).all()

    def _is_role_in_plans(self, role_plans, role) -> bool:
        for plan in role_plans:
            plan_role_type = RoleType("collection:" + plan.doc["role"])
            if plan.doc["principal"] == role.principal and plan_role_type == role.role:
                return True
        return False

    def _create_role(self, doc: Dict[str, Any], collection_id: str) -> models.Role:
        role = models.Role(doc={**doc, "collection": collection_id})
        create_object(role)
        return role

    def _delete_role(self, role: models.Role) -> None:
        if role.primary_role:
            click.echo(
                f"Cannot delete primary role {role.id} on collection {role.resource}",
                err=True,
            )
            return
        delete_object(role)

    def _check_globus_environment(self, app_env: str, plan_env: Optional[str]) -> None:
        if plan_env == "default":
            plan_env = "production"

        if plan_env is None:
            if app_env in ["default", "production"]:
                return
            click.echo(
                "The migration plan was created for the production environment, it will not work with "
                f"an endpoint in the {app_env} environment.",
                err=True,
            )
            click.get_current_context().exit(1)

        if app_env.upper() != plan_env.upper():
            click.echo(
                f"The migration plan was created for the {plan_env} environment, it will not work with "
                f"an endpoint in the {app_env} environment.",
                err=True,
            )
            click.get_current_context().exit(1)

    def _migration_display_name(self, display_name: str) -> str:
        return self.COLLECTION_DISPLAY_NAME_FORMAT.format(display_name)

    def _migration_keyword(self, v4_collection_id: str) -> str:
        return self.COLLECTION_KEYWORD_FORMAT.format(v4_collection_id)

    def _insert_migration_keyword(
        self, keywords: List[str], v4_collection_id: str
    ) -> List[str]:
        return keywords + [self._migration_keyword(v4_collection_id)]

    def _get_id_from_migration_keywords(self, keywords: List[str]) -> Optional[str]:
        """
        If keywords contains a migration keyword, return the ID from the keyword.
        """
        uuid_len = 36
        uuid_regex = (
            "[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-f-F]{4}-[0-9a-fA-F]{12}"
        )
        keyword_regex = self._migration_keyword(uuid_regex)
        pattern = f"^{keyword_regex}$"

        for kw in keywords:
            if re.search(pattern, kw):
                return kw[-uuid_len:]
        return None

    # Used throughout the following comparison functions
    MATCH = 0
    NO_MATCH = 1

    @classmethod
    def _cmp_path_restrictions(
        cls, model_value: PathRestrictions, doc_value: Dict[str, Any]
    ) -> int:
        """
        Compare path_restrictions between the existing model and the migration doc.
        Returns 0 if values match, non-zero otherwise.
        """
        for field in ["read", "read_write", "none"]:
            if set(getattr(model_value, field)) != set(doc_value.get(field, [])):
                return cls.NO_MATCH
        return cls.MATCH

    @classmethod
    def _cmp_sharing_restrict_paths(
        cls, model_value: PathRestrictions, doc_value: Optional[Dict[str, Any]]
    ) -> int:
        """
        Compare mapped collection sharing_restrict_paths between the existing model
        and the migration doc. Returns 0 if values match, non-zero otherwise.

        The default sharing_restrict_paths value for the model is:
            {'_read': [], '_read_write': ['/'], '_none': []}
        """
        default_sharing_restrict_paths = {"read": [], "read_write": ["/"], "none": []}
        if doc_value is not None:
            return cls._cmp_path_restrictions(model_value, doc_value)
        return cls._cmp_path_restrictions(model_value, default_sharing_restrict_paths)

    @classmethod
    def _cmp_restrict_paths(
        cls,
        model_value: Optional[PathRestrictions],
        doc_value: Optional[Dict[str, Any]],
    ) -> int:
        """
        Compare storage gateway restrict_paths between the existing model
        and the migration doc. Returns 0 if values match, non-zero otherwise.

        The default restrict_paths value for the model is 'None'.
        """
        if model_value is None and doc_value is None:
            return cls.MATCH
        if model_value is None or doc_value is None:
            return cls.NO_MATCH
        return cls._cmp_path_restrictions(model_value, doc_value)

    IdentityMappingArg1 = TypeVar("IdentityMappingArg1")
    IdentityMappingArg2 = TypeVar("IdentityMappingArg2")

    @classmethod
    def _cmp_mapping_loop(
        cls,
        outer_list: List[IdentityMappingArg1],
        inner_list: List[IdentityMappingArg2],
        cmp_func: Callable[[IdentityMappingArg1, IdentityMappingArg2], int],
    ) -> int:
        """
        Try to find every item in outer_list in inner_list.
        """
        for outer_val in outer_list:
            found_match = False
            for inner_val in inner_list:
                if cmp_func(outer_val, inner_val) == cls.MATCH:
                    found_match = True
                    break
            if not found_match:
                return cls.NO_MATCH
        return cls.MATCH

    @classmethod
    def _cmp_expression_mappings(
        cls, model_value: ExpressionIdentityMapping, doc_value: Dict[str, Any]
    ) -> int:
        def _cmp_expression_mapping(
            model_mapping: Union[CompiledExpressionIdentityMappingDict, Dict[str, Any]],
            doc_mapping: Union[CompiledExpressionIdentityMappingDict, Dict[str, Any]],
        ) -> int:
            for attr, default in [
                ("source", None),
                ("match", None),
                ("ignore_case", False),
                ("literal", False),
                ("output", None),
            ]:
                if model_mapping.get(attr, default) != doc_mapping.get(attr, default):
                    return cls.NO_MATCH
            return cls.MATCH

        rv = cls._cmp_mapping_loop(
            model_value.mappings,
            doc_value["mappings"],
            _cmp_expression_mapping,
        )
        if rv == cls.NO_MATCH:
            return cls.NO_MATCH

        return cls._cmp_mapping_loop(
            doc_value["mappings"],
            model_value.mappings,
            _cmp_expression_mapping,
        )

    @classmethod
    def _cmp_external_mapping(
        cls, model_value: ExternalIdentityMapping, doc_value: Dict[str, Any]
    ) -> int:
        if model_value.command != doc_value["command"]:
            return cls.NO_MATCH
        return cls.MATCH

    @classmethod
    def _cmp_identity_mappings(
        cls,
        model_value: Optional[
            List[Union[ExpressionIdentityMapping, ExternalIdentityMapping]]
        ],
        doc_value: Optional[List[Dict[str, Any]]],
    ) -> int:
        """
        Compare the identity_mappings on the storage gateway to the mappings supplied in
        the migration plan. Return MATCH if the two are equivalent, return NO_MATCH otherwise.
        """
        if model_value is None and doc_value is None:
            return cls.MATCH
        if model_value is None or doc_value is None:
            return cls.NO_MATCH

        def _cmp_identity_mapping(model_mapping, doc_mapping):
            """
            Compare two mappings which might not have the same DATA_TYPE.
            """
            if model_mapping.DATA_TYPE != doc_mapping["DATA_TYPE"]:
                return cls.NO_MATCH
            mapping_type = model_mapping.DATA_TYPE.split("#")[0]
            if (
                mapping_type == "expression_identity_mapping"
                and cls._cmp_expression_mappings(model_mapping, doc_mapping)
                == cls.MATCH
            ):
                return cls.MATCH

            if (
                mapping_type == "external_identity_mapping"
                and cls._cmp_external_mapping(model_mapping, doc_mapping) == cls.MATCH
            ):
                return cls.MATCH

            return cls.NO_MATCH

        rv = cls._cmp_mapping_loop(
            model_value,
            doc_value,
            _cmp_identity_mapping,
        )
        if rv == cls.NO_MATCH:
            return cls.NO_MATCH

        return cls._cmp_mapping_loop(
            doc_value, model_value, lambda x, y: _cmp_identity_mapping(y, x)
        )


def _import_migration_plan() -> MigrationPlan:
    """
    Load the GCS migration plan from disk.
    """
    plan_sub_directory = "migration_plan"

    if not os.path.isdir(plan_sub_directory):
        click.echo(
            f"{plan_sub_directory} not found in the current directory. "
            "Change to the directory containing the plan and rerun this command.",
            err=True,
        )
        click.get_current_context().exit(1)

    try:
        migration_plan = MigrationPlan.load()
    except KeyError as e:
        click.echo(
            f"Failed to load the migration plan because of an internal issue. "
            f'Please report this error to support@globus.org: \n"KeyError: {e}"',
            err=True,
        )
        click.get_current_context().exit(1)

    return migration_plan


def _check_imported_plan(migration_plan: MigrationPlan) -> None:
    """
    Run gcs_migration checks on on all components in the plan.
    """

    T = TypeVar("T")
    WalkReturnType = Tuple[Type[check.CheckInvoker[T]], Optional[T]]

    def _walk_plan(migration_plan: MigrationPlan) -> Iterator[WalkReturnType]:
        yield (check.MigrationPlanCheck, migration_plan)
        yield (
            check.StorageGatewayCheck,
            migration_plan.storage_gateway,
        )
        yield (
            check.MappedCollectionCheck,
            migration_plan.mapped_collection,
        )

        assert migration_plan.mapped_collection is not None
        for role in cast(
            List[migration_models.Role], migration_plan.mapped_collection.roles
        ):
            yield (check.RoleCheck, role)

        for guest in migration_plan.guest_collections or []:
            yield (check.GuestCollectionCheck, guest)
            yield (check.CredentialCheck, guest.credential)

            for role in guest.roles or []:
                yield (check.RoleCheck, role)

            if guest.acls is not None:
                for acl in guest.acls:
                    yield (check.ACLCheck, acl)

    for t in _walk_plan(migration_plan):
        res = t[0]().check(t[1])
        if res:
            click.echo(
                "The migration plan is not ready. Run 'gcs-migration check' and "
                "fix all issues:\n    "
                + "\n    ".join([r.diagnostic or "" for r in res]),
                err=True,
            )
            click.get_current_context().exit(1)


class ModelToJSON(Protocol):
    def to_json(self):
        ...


def _load_doc(
    name: str, model: ModelToJSON, schema: Optional[ma.Schema]
) -> Dict[str, Any]:
    """
    Helper function that wraps these calls in order to print useful information
    on exception.

    Note: Importing the the to_json() docs via our schema will transform some fields
    to make them suitable for use with our models. For example. collection.collection_type
    transforms from a string ('mapped', 'guest') to an enum.
    """
    try:
        doc_in = model.to_json()
    except AttributeError as e:
        click.echo(
            f"Failed to dump {name} document:\n"
            f"{str(e)}\n"
            "This is an internal error that needs to be reported to support@globus.org",
            err=True,
        )
        click.get_current_context().exit(1)

    try:
        if schema:
            doc_out = schema.load(doc_in)
        else:
            doc_out = doc_in
    except ma.ValidationError as e:
        click.echo(
            f"Failed to load {name} document:\n"
            f"{str(e)}\n"
            "This is an internal error that needs to be reported to support@globus.org",
            err=True,
        )
        click.get_current_context().exit(1)

    return doc_out
