import datetime
import pwd
from typing import Any, Dict, List, Optional

import globus_sdk

import globus.manager.api.models as models
from globus.manager.api.schemas.api import user_credential_private_schema


class _UserCredentialV53:
    """
    Base class of a storage gateway exported from a GCSv5.3 installation. This
    maps the properties of the storage gateway into something which can be
    dumped using the v5.4 marshmallow schema. It also contains some private
    fields which do not directly map to v5.4 but will be needed to figure out
    what types of storage gateways will be needed to process the collections
    on this storage gateway.
    """

    model_class = models.UserCredential

    def __init__(self, auth_client: globus_sdk.ConfidentialAppAuthClient, **kwargs):
        self._kwargs = kwargs
        self._auth_client = auth_client
        self.identity_id = kwargs.get("identity_id")
        self.connector_id = kwargs.get("connector_id")
        self.storage_gateway_id = kwargs["storage_gateway_id"]
        self.username = kwargs.get("storage_gateway_user_id")
        self.dummy_credential = self.username == ":globus-mapping:"
        self.display_name = kwargs["display_name"]

        self.deleted = kwargs.get("deleted", False)
        self.invalid = kwargs.get("invalid", False)
        self.provisioned = True

        # These are stored, but are only used internally and are not visible
        # to the outside
        self._created_at = kwargs["created_at"]
        self._updated_at = kwargs["updated_at"]

        storage_credential = kwargs.get("storage_credential")
        self.init_policies(storage_credential)

        self.id = models.UserCredential.id_for_account(
            models.Account(
                username=self.username,
                storage_gateway_id=self.storage_gateway_id,
                identity_id=self.identity_id,
            )
        )

    def init_policies(self, credential):
        pass

    def _resolve_identity(self, identity_id):
        ids = self._auth_client.get_identities(ids=[identity_id])
        return ids["identities"][0]["username"]


class PosixUserCredentialV53(_UserCredentialV53):
    model_class = models.PosixUserCredential

    def init_policies(self, credential):
        if self.username != ":globus-mapping:":
            if credential is not None:
                assert credential == self.username
            else:
                credential = self.username
            try:
                self.display_name = pwd.getpwnam(self.username).pw_gecos.split(",")[0]
            except KeyError:
                self.display_name = "DELETED USER"
        else:
            self.display_name = self.username
        super().init_policies(credential)


class GoogleCloudStorageUserCredentialV53(_UserCredentialV53):
    model_class = models.GoogleCloudStorageUserCredential

    def init_policies(self, credential: Optional[Dict[str, Any]]):
        if credential is None:
            credential = {}
        self._init_policies(**credential)

    def _init_policies(
        self,
        email: str = None,
        refresh_token: str = None,
        scopes: List[str] = None,
        **kwargs
    ):
        self.sub = self.username
        if email:
            self.email = email
            self.username = email
        else:
            self.email = self.display_name
            self.username = self.display_name

        if refresh_token:
            self.refresh_token = refresh_token
            self.token_expiry = datetime.datetime.now(tz=datetime.timezone.utc)
        if scopes:
            self.scopes = scopes

        super().init_policies(kwargs)


class GoogleDriveUserCredentialV53(_UserCredentialV53):
    model_class = models.GoogleDriveUserCredential

    def init_policies(self, credential: Optional[Dict[str, Any]]):
        if credential is None:
            credential = {}
        self._init_policies(**credential)

    def _init_policies(
        self,
        email: str = None,
        refresh_token: str = None,
        scopes: List[str] = None,
        **kwargs
    ):
        self.sub = self.username
        if email:
            self.email = email
            self.username = email
        else:
            self.email = self.display_name
            self.username = self.display_name

        if refresh_token:
            self.refresh_token = refresh_token
            self.token_expiry = datetime.datetime.now(tz=datetime.timezone.utc)
        if scopes:
            self.scopes = scopes

        super().init_policies(kwargs)


class S3UserCredentialV53(_UserCredentialV53):
    model_class = models.S3UserCredential

    def init_policies(self, credential: Optional[Dict[str, Any]]):
        if credential is None:
            credential = {}
        self._init_policies(**credential)

    def _init_policies(
        self, s3_key_id: str = None, s3_secret_key: str = None, **kwargs
    ):
        if s3_key_id:
            self.s3_key_id = s3_key_id
        if s3_secret_key:
            self.s3_secret_key = s3_secret_key
        self.username = self._resolve_identity(self.identity_id)
        super().init_policies(kwargs)


class CephUserCredentialV53(_UserCredentialV53):
    model_class = models.CephUserCredential

    def init_policies(self, credential: str = None, **kwargs):
        if credential is not None:
            assert credential == self.username
        else:
            credential = self.username
        super().init_policies(kwargs)


class BoxUserCredentialV53(_UserCredentialV53):
    model_class = models.BoxUserCredential

    def init_policies(self, credential: str = None, **kwargs):
        if credential is not None:
            assert credential == self.username
        else:
            credential = self.username
        super().init_policies(kwargs)


class BlackPearlUserCredentialV53(_UserCredentialV53):
    model_class = models.BlackPearlUserCredential

    def init_policies(self, credential: str = None, **kwargs):
        if credential is not None:
            assert credential == self.username
        else:
            credential = self.username
        super().init_policies(kwargs)


def UserCredentialV53(
    storage_gateway: models.StorageGateway,
    auth_client: globus_sdk.ConfidentialAppAuthClient,
    **kwargs
) -> models.UserCredential:
    """
    Pseudo-constructor for the *UserCredentialV53 classes, which uses the
    connector id from the storage gateway definition document to select which
    type of object to create.
    """

    connector_map = {
        "7e3f3f5e-350c-4717-891a-2f451c24b0d4": BlackPearlUserCredentialV53,
        "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63": BoxUserCredentialV53,
        "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72": CephUserCredentialV53,
        "56366b96-ac98-11e9-abac-9cb6d0d9fd63": GoogleCloudStorageUserCredentialV53,
        "976cf0cf-78c3-4aab-82d2-7c16adbcc281": GoogleDriveUserCredentialV53,
        "145812c8-decc-41f1-83cf-bb2a85a2a70b": PosixUserCredentialV53,
        "7643e831-5f6c-4b47-a07f-8ee90f401d23": S3UserCredentialV53,
    }

    connector_id = storage_gateway.connector_id
    assert connector_id in connector_map

    uc = connector_map[connector_id](
        connector_id=connector_id, auth_client=auth_client, **kwargs
    )
    return uc.model_class(
        appsync_data={
            "id": uc.id,
            "configuration_hash": user_credential_private_schema.dump(uc),
        },
        storage_gateway=storage_gateway,
    )
