import os
import ipaddress
import pwd
from typing import Any, Optional, Tuple, Union


from distro import linux_distribution
from platform import architecture

from ..service_manager import ServiceManager
from ..utils import (
    HostInfo,
    log,
    write_and_link_file,
)
from globus.manager.version import __version__ as version


PRIVATE_DATA_WARNING = """
******************************************************************************
WARNING: Your GridFTP server is behind a NAT, but the Server name resolves
to a private IP address. This probably won't work correctly with Globus.
******************************************************************************
"""

TRANSFER_ADDR = "54.237.254.192/29"


class GridFTPService:
    service = "globus-gridftp-server"
    """
    Class to configure a GridFTP server
    """

    def __init__(
        self,
        incoming_port_range: Optional[Tuple[int, int]],
        outgoing_port_range: Optional[Tuple[int, int]],
        data_interface: Optional[Union[ipaddress.IPv4Address, ipaddress.IPv6Address]],
        domain: Optional[str],
        config_owner: str = "gcsweb",
        allow_udt: bool = False,
        environment: str = "default",
        require_encryption: bool = False,
        extra_arg: Optional[str] = None,
        service_manager: Any = None,
        root: str = "/",
    ) -> None:
        self.server = domain
        if service_manager is None:
            service_manager = ServiceManager
        self.service_manager = service_manager
        self.ipc_uds_path = os.path.join(root, "var/run/globus-connect-server/ipc")
        self.control_uds_path = os.path.join(
            root, "var/run/globus-connect-server/control"
        )
        self.trusted_ca_dir = os.path.join(
            root, "var/lib/globus-connect-server/grid-security/certificates"
        )
        self.vhost_dir = os.path.join(root, os.path.expanduser("~gcsweb/etc/domains"))
        self.etc_gridftp_d = os.path.join(root, "etc/gridftp.d")
        self.var_gridftp_d = os.path.join(
            root, "var/lib/globus-connect-server/gridftp.d"
        )
        self.sharing_dn = "/".join(
            [
                "",
                "C=US",
                "O=Globus Consortium",
                "OU=Globus Online",
                "OU=Transfer User",
                "CN=__transfer__",
            ]
        )
        self.incoming_range = incoming_port_range if incoming_port_range else None
        self.outgoing_range = outgoing_port_range if outgoing_port_range else None
        self.config_owner = config_owner
        self.allow_udt = allow_udt
        self.gridftp_keyfile = os.path.expanduser("~gcsweb/gridftp-key")
        self.config_uri = "https://{}/api/private/gridftp_login".format(domain)
        self.config_unix_socket = "/run/gcs_manager.sock"
        self.allow_from = None
        if environment.lower() == "production" or environment.lower() == "default":
            self.allow_from = TRANSFER_ADDR
        self.data_interface = str(data_interface) if data_interface else None
        self.require_encryption = require_encryption
        self.extra_arg = extra_arg

    def update(self, force: bool = False) -> None:
        if HostInfo.is_private_ip(self.data_interface):
            log.warn(PRIVATE_DATA_WARNING)

        conf_file_name, conf_link_name = self.__config_names()
        config_options = {
            "auth_level 19": True,
            "port_range {0[0]},{0[1]}": self.incoming_range,
            "$GLOBUS_TCP_SOURCE_RANGE {0[0]},{0[1]}": self.outgoing_range,
            "data_interface {}": self.data_interface,
            "combined_data_node 1": True,
            "ipc_uds_path {}": self.ipc_uds_path,
            "control_uds_path {}": self.control_uds_path,
            "ipc_port {}": -1,
            "control_tls 1": True,
            "allow_udt 1": self.allow_udt or None,
            "allow_from {}": self.allow_from,
            "threads 1": True,
            "encrypt_data 1": self.require_encryption or None,
            "config_keyfile {}": self.gridftp_keyfile,
            "config_uri {}": self.config_uri,
            "config_unix_socket {}": self.config_unix_socket,
            'sharing_dn "{}"': self.sharing_dn,
            '$X509_CERT_DIR "{}"': self.trusted_ca_dir,
            '$X509_VHOST_CRED_DIR "{}"': self.vhost_dir,
            "log_audit {}": "/var/log/gridftp-audit.log",
            "log_single {}": "/var/log/gridftp.log",
            "log_level {}": "ERROR,WARN",
            "{}": self.extra_arg,
        }
        write_and_link_file(
            conf_file_name,
            conf_link_name,
            "\n".join(
                [
                    opt.format(arg)
                    for opt, arg in config_options.items()
                    if arg is not None
                ]
                + [""]
            ),
        )

    def setup(self, force=False):
        self.update(force)

        if self.config_owner:
            pwent = pwd.getpwnam(self.config_owner)
            os.chown(self.__config_names()[0], pwent.pw_uid, pwent.pw_gid)

        self.service_manager.enable(self.service)
        if self.service_manager.status(self.service):
            self.service_manager.restart(self.service)
        else:
            self.service_manager.start(self.service)

    @classmethod
    def usage_stats_id(cls):
        dist_string = "-".join(
            [
                d
                for d in list(linux_distribution(full_distribution_name=False))[0:2]
                if d != ""
            ]
            + [a for a in list(architecture())[0:1] if a != ""]
        )
        usid = "GCS-{}".format(version)
        if dist_string != "":
            usid += "+{}".format(dist_string)
        return usid

    def cleanup(self, **kwargs):
        self.service_manager.stop(self.service)
        self.service_manager.disable(self.service)

        for name in os.listdir(self.etc_gridftp_d):
            if (
                name.startswith("globus-connect-server")
                and name != "globus-connect-server-package"
            ):
                os.remove(os.path.join(self.etc_gridftp_d, name))
        for name in os.listdir(self.var_gridftp_d):
            os.remove(os.path.join(self.var_gridftp_d, name))

    def __config_names(self, name=None, clean_link=True):
        if name is not None:
            longname = "globus-connect-server-" + name
        else:
            longname = "globus-connect-server"

        conf_file_name = os.path.join(self.var_gridftp_d, longname)
        conf_link_name = os.path.join(self.etc_gridftp_d, longname)
        return (conf_file_name, conf_link_name)
