import logging
import os
from base64 import b64encode

from flask import session
import globus_oidc.pam_helper as pam
from mako.lookup import TemplateLookup

from oic.utils.authn.user import UsernamePasswordMako, UserAuthnMethod
from globus_oidc.config import settings

log = logging.getLogger(__name__)

LOOKUP = TemplateLookup(directories=[settings()["site"]["dir"]],
                        input_encoding='utf-8', output_encoding='utf-8')


class UsernamePasswordPAM(UsernamePasswordMako):

    # Cookie lifetime. Needs to be sufficiently large to account for users
    # whose machine clocks are off by non-trivial amounts and who are also
    # using a browser that depends solely on the local clock.
    #
    # Beyond that this won't matter because the authorization view will unset
    # the login cookie.
    ttl = 24*60

    def __init__(self, srv, mako_template, template_lookup, pwd, return_to="",
                 templ_arg_func=None, verification_endpoints=None):
        # Overriding parent __init__ to let us set ttl here:
        UserAuthnMethod.__init__(self, srv, ttl=self.ttl)
        self.mako_template = mako_template
        self.template_lookup = template_lookup
        self.passwd = pwd
        self.return_to = return_to
        self.verification_endpoints = verification_endpoints or ["verify"]
        # PAM service.
        self.pam_service = settings()['config']['oidc_server']['pam_service']
        self.support_email = settings()['config']['oidc_server']['support_email']
        self.support_contact = settings()['config']['oidc_server']['support_contact']
        self.display_name = settings()['config']['oidc_server']['display_name']

        if templ_arg_func:
            self.templ_arg_func = templ_arg_func
        else:
            self.templ_arg_func = self.template_args

    def template_args(self, end_point_index=0, **kwargs):
        argv = super(UsernamePasswordPAM, self).template_args(end_point_index, **kwargs)
        # Set the CSRF in the form data
        argv['csrf'] = b64encode(os.urandom(15)).decode()
        # Set the same CSRF in the session.
        session['csrf'] = argv['csrf']
        argv['support_email'] = self.support_email
        argv['support_contact'] = self.support_contact
        argv['display_name'] = self.display_name
        return argv

    def _verify(self, password: str, username: str):
        """
        Returns the result of authenticating the user+password against the local login service.
        Note: If you are unable to log in users other than yourself,
        change the permissions of the running file with:
        `chmod u+s [file_that_is_running_this_program]`.
        :param user: The username
        :param password: The password for the user
        :return: Nothing if authenticated or an AssertionError if not.
        """
        UsernamePasswordPAM.authenticate(self.pam_service, username, password)

    @staticmethod
    def authenticate(pam_service: str, username: str, password: str):
        p = pam.pam()
        res = p.authenticate(username, password, pam_service)
        log.debug("User '{}', result: {} - {} {}".format(username, res, p.code, p.reason))

        if not res:
            raise AssertionError()  # Fail
        return


def pam_authentication(base_url):
    return UsernamePasswordPAM(
                        None,
                        "login.mako",
                        LOOKUP,
                        # Password database, not used since we're using PAM-Kerberos
                        None,
                        return_to="{}/authorization".format(base_url))
